#!/usr/bin/env python

"""Classes that provide abstractions of different image source manifests."""

import abc
import copy
import json

from pathlib import Path
from typing import List

from .imageconfig import ImageConfig
from .imagename import ImageName
from .utils import formatted_digest, read_file, FormattedSHA256


class Manifest(abc.ABC):
    """
    Abstract class to retrieve and manipulate image source manifests.
    """

    def __init__(self, manifest: bytes):
        """
        Args:
            manifest: The raw image source manifest value.
        """
        self.json = json.loads(manifest)

    def __str__(self):
        return json.dumps(self.json)

    @abc.abstractmethod
    def get_config_digest(self, image_name: ImageName = None) -> FormattedSHA256:
        """
        Retrieves the image configuration digest value from the image source manifest.

        Args:
            image_name: The image name.

        Returns:
            The image configuration digest value in the form: <hash type>:<digest value>.
        """

    @abc.abstractmethod
    def get_layers(self, image_name: ImageName = None) -> List[FormattedSHA256]:
        """
        Retrieves the listing of manifest layer identifiers.

        Args:
            image_name: The image name.

        Returns:
            list: Layer identifiers in the form: <hash type>:<digest value>.
        """

    # TODO: Move weird, type-specific, manifest logic to these methods
    # @abc.abstractmethod
    # def copyfrom(self, src_image_source: ImageSource, dest_image_name, config_digest_signed) -> str:


class ArchiveManifest(Manifest):
    """
    Image source manifest for docker archives.
    """

    @staticmethod
    def layer_to_digest(layer: str) -> FormattedSHA256:
        """
        Coverts a archive layer identifier to a digest value.

        Args:
            layer: The archive layer identifier (relative tar path).

        Returns:
            The corresponding digest value in the form: <hash type>:<digest value>.
        """
        return FormattedSHA256(layer[:-10])

    @staticmethod
    def digest_to_layer(digest: FormattedSHA256) -> str:
        """
        Converts a digest value to a archive layer identifier.

        Args:
            digest: The digest value in the form: <hash type>:<digest value>.

        Returns:
            The corresponding archive layer identifier (relative tar path).
        """
        return "{0}/layer.tar".format(digest.sha256)

    def get_config(self, image_name: ImageName):
        """
        Retrieves the image configuration dictionary for a given image name from the image source manifest.

        Args:
            image_name: The image name.

        Returns:
            A dictionary as defined by the Docker specification.
            dict:
                Config: Image configuration (relative tar path).
                RepoTags: Docker repository tags.
                Layers: List of image layer (relative tar paths).
        """
        if image_name.tag:
            repotag = str(image_name)
            for config in self.json:
                if config["RepoTags"] and repotag in config["RepoTags"]:
                    return config
        else:
            return [
                config
                for config in self.json
                if config["Config"] == "{0}.json".format(image_name.image)
            ][0]
        raise RuntimeError("Unable to locate configuration in archive manifest!")

    def append_config(
        self,
        config_digest: FormattedSHA256,
        layers: List[FormattedSHA256],
        repotags: List = None,
    ):
        """
        Appends an image configuration dictionary to the image source manifest

        Args:
            config_digest: Image configuration digest in the form <hash type>:<digest value>.
            layers: List of image layer identifiers in the form: <hash type>:<digest_value>.
            repotags: List of docker repository tags.
        """
        config = {
            "Config": "{0}.json".format(config_digest.sha256),
            "Layers": [ArchiveManifest.digest_to_layer(l) for l in layers],
        }
        if repotags:
            config["RepoTags"] = repotags

        # TODO: Check to make sure it doesn't already exist?
        self.json.append(config)

    # Manifest Members

    def get_config_digest(self, image_name: ImageName = None) -> FormattedSHA256:
        config = self.get_config(image_name)
        return FormattedSHA256(config["Config"][:-5])

    def get_layers(self, image_name: ImageName = None) -> List[FormattedSHA256]:
        layers = self.get_config(image_name)["Layers"]
        return [ArchiveManifest.layer_to_digest(l) for l in layers]


class RegistryV2Manifest(Manifest):
    """
    Manifest list, aka "fat manifest" as defined in:

    https://github.com/docker/distribution/blob/master/docs/spec/manifest-v2-2.md
    """

    def set_config_digest(self, config_digest: FormattedSHA256, size: int):
        """
        Assigns the image configuration digest and size in the image source manifest.

        Args:
            config_digest: Image configuration digest in the form <hash type>:<digest value>.
            size: Image configuration size in bytes.
        """
        if self.json["schemaVersion"] == 2:
            self.json["config"]["digest"] = config_digest
            self.json["config"]["size"] = size
        else:
            raise RuntimeError(
                "Unsupported schema version: {0}".format(self.json["schemaVersion"])
            )

    def set_layers(self, layers: List[FormattedSHA256]):
        """
        Assigns the list of manifest layer identifiers.

        Note: It is not the intention of this utility to modify layer content! As such we only support updating the
              manifest layer identifiers as part of layer replication between two disjoint registries. (i.e.
              Modification of the overall layer count, order, or individual layer sizes is explicitly not implemented.)

        Args:
            layers: List of manifest layer identifiers in the form: <hash type>:<digest_value>.
        """
        if self.json["schemaVersion"] == 2:
            for i, layer in enumerate(self.json["layers"]):
                layer["digest"] = layers[i]
        else:
            raise RuntimeError(
                "Unsupported schema version: {0}".format(self.json["schemaVersion"])
            )

    # Manifest Members

    def get_config_digest(self, image_name: ImageName = None) -> FormattedSHA256:
        if self.json["schemaVersion"] == 2:
            result = FormattedSHA256.parse(self.json["config"]["digest"])
        else:
            raise RuntimeError(
                "Unsupported schema version: {0}".format(self.json["schemaVersion"])
            )

        return result

    def get_layers(self, image_name: ImageName = None) -> List[FormattedSHA256]:
        if self.json["schemaVersion"] == 2:
            result = [FormattedSHA256.parse(l["digest"]) for l in self.json["layers"]]
        else:
            raise RuntimeError(
                "Unsupported schema version: {0}".format(self.json["schemaVersion"])
            )

        return result


class DeviceMapperRepositoryManifest(Manifest):
    """
    Image source manifest for docker repositories.
    """

    @staticmethod
    def _get_repository_key(image_name: ImageName) -> str:
        """
        Retrieves the repository key for a given image name.

        Args:
            image_name: Image name for which to retrieve the key.

        Returns:
            The corresponding repository key.
        """
        key_name = copy.deepcopy(image_name)
        key_name.tag = None
        return str(key_name)

    @staticmethod
    def get_combined_layerid(
        parent: FormattedSHA256, layer: FormattedSHA256
    ) -> FormattedSHA256:
        """
        Retrieves the layer identifier for a given parent-layer combination.

        Args:
            parent: The parent layer identifier.
            layer:  The layer identifier.

        Returns:
            The corresponding layer identifier.
        """
        result = layer
        if parent:
            # Note: The image layer is the digest value of the formatted string.
            result = formatted_digest("{0} {1}".format(parent, layer).encode("utf-8"))
        return result

    def override_config(self, config_digest: FormattedSHA256, image_name: ImageName):
        """
        Assigns the image configuration digest for a given image name in the image source manifest.

        Args:
            config_digest: Image configuration digest in the form <hash type>:<digest value>.
            image_name: The image name.
        """

        # Note: registry.json can only contain two formats:
        #       [<endpoint>/]<image>:<tag>
        #       [<endpoint>/]<image>@<hash type>:<digest value>
        key = DeviceMapperRepositoryManifest._get_repository_key(image_name)
        image = self.json["Repositories"].get(key, {})
        image[str(image_name)] = config_digest
        self.json["Repositories"][key] = image

    # Manifest Members

    def get_config_digest(self, image_name: ImageName = None) -> FormattedSHA256:
        key = DeviceMapperRepositoryManifest._get_repository_key(image_name)
        image = self.json["Repositories"].get(key, {})
        return FormattedSHA256.parse(image.get(str(image_name), None))

    def get_layers(self, image_name: ImageName = None) -> List[FormattedSHA256]:
        # TODO: How can this be reimplemented to avoid duplicate code?
        DM_CONTENT_ROOT = Path(
            "/var/lib/docker/image/devicemapper/imagedb/content/sha256"
        )
        path = DM_CONTENT_ROOT.joinpath(self.get_config_digest(image_name).sha256)
        image_config = ImageConfig(read_file(path))

        result = []
        parent = None
        for layer in image_config.get_image_layers():
            result.append(
                DeviceMapperRepositoryManifest.get_combined_layerid(parent, layer)
            )
            parent = result[-1]

        return result



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""ImageSource tests."""

import json
import pytest

from docker_sign_verify import ImageName, ImageSource, Signer

from .stubs import _signer_for_signature, FakeRegistryV2ImageSourceNoLabels
from .test_imagename import image_name, url

# TODO: Additionally, test all methods with config.json that *does* contain 'Labels' ...


@pytest.fixture
def fake_registry_v2_image_source(request) -> FakeRegistryV2ImageSourceNoLabels:
    """Provides a fake RegistryV2ImageSource without"""
    # Do not use caching; get a new instance for each test
    return FakeRegistryV2ImageSourceNoLabels(request, layer_exists=True, dry_run=True)


def test_init(fake_registry_v2_image_source: ImageSource):
    """Test that the image source can be instantiated."""
    assert fake_registry_v2_image_source


def test__sign_image_config(
    fake_registry_v2_image_source: FakeRegistryV2ImageSourceNoLabels,
    image_name: ImageName,
):
    """Test adding signature(s) to the image configuration."""
    result = fake_registry_v2_image_source.quick_sign(image_name)
    assert result

    image_config = result["image_config"]
    assert image_config
    assert "FAKE SIGNATURE" in str(image_config)
    assert json.loads(image_config.get_config())

    signature_value = result["signature_value"]
    assert signature_value
    assert "FAKE SIGNATURE" in signature_value

    verify_image_data = result["verify_image_data"]
    assert verify_image_data
    assert image_config == verify_image_data["image_config"]

    manifest = verify_image_data["manifest"]
    assert manifest
    assert manifest.get_config_digest() == image_config.get_config_digest()
    assert len(manifest.get_layers()) == len(image_config.get_image_layers())


def test__verify_image_config(
    fake_registry_v2_image_source: FakeRegistryV2ImageSourceNoLabels,
    image_name: ImageName,
):
    """Test verifying the integrity of the image configuration."""

    def assertions(result: dict):
        assert result

        image_config = result["image_config"]
        assert image_config
        assert json.loads(image_config.get_config())

        image_layers = result["image_layers"]
        assert image_layers

        manifest = result["manifest"]
        assert manifest
        assert manifest.get_config_digest() == image_config.get_config_digest()
        assert json.loads(image_config.get_config())

        manifest_layers = result["manifest_layers"]
        assert manifest_layers
        assert len(image_layers) == len(manifest_layers)

    # 1. Pre signature
    # pylint: disable=protected-access
    assertions(fake_registry_v2_image_source._verify_image_config(image_name))

    # Sign
    fake_registry_v2_image_source.quick_sign(image_name)

    # 2. Post signature
    # pylint: disable=protected-access
    assertions(fake_registry_v2_image_source._verify_image_config(image_name))


def test_verify_image_signatures(
    fake_registry_v2_image_source: FakeRegistryV2ImageSourceNoLabels,
    image_name: ImageName,
):
    """Test verifying the signatures within the image configuration."""
    # An exception should be raised if the image configuration is not signed
    with pytest.raises(Exception) as exception:
        fake_registry_v2_image_source.verify_image_signatures(image_name)
    assert str(exception.value) == "Image does not contain any signatures!"

    # Sign
    fake_registry_v2_image_source.quick_sign(image_name)

    # Replace the class method for resolving signature providers ...
    original_method = Signer.for_signature
    Signer.for_signature = _signer_for_signature

    result = fake_registry_v2_image_source.verify_image_signatures(image_name)
    assert result["image_config"]
    assert result["signatures"]

    # Restore the original class method
    Signer.for_signature = original_method



#!/usr/bin/env python

"""Utility classes."""

import gzip
import hashlib
import io
import os
import shutil
import tarfile
import tempfile
import time

from typing import Dict

CHUNK_SIZE = 4096


class FormattedSHA256(str):
    """A algorithm prefixed SHA256 hash value."""

    def __new__(cls, sha256: str):
        if not sha256 or len(sha256) != 64:
            raise ValueError(sha256)
        obj = super().__new__(cls, "sha256:{0}".format(sha256))
        obj.sha256 = sha256
        return obj

    @staticmethod
    def parse(digest: str):
        """
        Initializes a FormattedSHA256 from a given SHA256 digest value.

        Args:
            digest: A SHA256 digest value in form SHA256:<digest value>.

        Returns:
            The newly initialized object.
        """
        if not digest or not digest.startswith("sha256:") or len(digest) != 71:
            raise ValueError(digest)
        return FormattedSHA256(digest[7:])


def _chunk_file(file_in, file_out) -> Dict:
    """
    Copies from one file to another in chunks.

    Args:
        file_in: The input file.
        file_out: The output file.

    Returns:
        dict:
            digest: The SHA256 digest value of the output file.
            size: Size of the output file in bytes.
    """
    size = 0
    hasher = hashlib.sha256()
    while True:
        chunk = file_in.read(CHUNK_SIZE)
        if not chunk:
            break
        file_out.write(chunk)
        size += len(chunk)
        hasher.update(chunk)
    file_out.flush()

    # Not all file object are native ...
    try:
        os.fsync(file_out.fileno())
    # pylint: disable=bare-except
    except:
        ...

    # Be kind, rewind ...
    file_out.seek(0)

    return {"digest": "sha256:{0}".format(hasher.hexdigest()), "size": size}


def must_be_equal(
    expected, actual, msg: str = "Actual value does not match expected value"
):
    """
    Compares two values and raises an exception if they are not equal.

    Args:
        expected: The expected value.
        actual: The actual value.
        msg: Message describing the context of the comparison.
    """
    if actual != expected:
        raise RuntimeError("{0}: {1} != {2}".format(msg, actual, expected))


def xellipsis(string: str) -> str:
    """
    Reduces the length of a given string, if it is wider than the terminal width, inserting str_ellipsis.

    Args:
        string: The string to be reduced.

    Returns:
        A string with a maximum length of the terminal width.
    """
    return str_ellipsis(string, os.environ.get("COLUMNS", 80) / 2)


def str_ellipsis(string: str, max_length: int = 40) -> str:
    """
    Reduces the length of a given string, if it is over a certain length, by inserting str_ellipsis.

    Args:
        string: The string to be reduced.
        max_length: The maximum length of the string.

    Returns:
        A string with a maximum length of :param:max_length.
    """
    if len(string) <= max_length:
        return string
    n_2 = int(max_length) / 2 - 3
    n_1 = max_length - n_2 - 3

    return "{0}...{1}".format(string[: int(n_1)], string[-int(n_2) :])


def formatted_digest(data: bytes) -> FormattedSHA256:
    """
    Returns the algorithm prefixed digest value of given data.

    Args:
        data: The data for which to calculate the digest value.

    Returns:
        The digest value of the data in the format <hast type>:<digest value>.
    """
    return FormattedSHA256(hashlib.sha256(data).hexdigest())


def copy_file(file_in, file_out) -> Dict:
    """
    Copies from one file to another.

    Args:
        file_in: The input file.
        file_out: The output file.

    Returns:
        dict: as defined by :func:~docker_sign_verify.Utils._chunk_file.
    """
    return _chunk_file(file_in, file_out)


def read_file(path) -> bytes:
    """
    Retrieves the entire contents of a file.

    Args:
        path: Absolute path of the file to be read.

    Returns:
        The file content.
    """
    bytesio = io.BytesIO()
    with path.open(mode="rb") as file:
        copy_file(file, bytesio)
    return bytesio.read()


def write_file(path, content: bytes) -> Dict:
    """
    Assigns the entire contents of a file.

    Args:
        path: Absolute path of the file to be assigned.
        content: The content to be assigned.

    Returns:
        dict: as defined by :func:~docker_sign_verify.Utils.copy_file.
    """
    bytesio = io.BytesIO(content)
    with path.open(mode="wb") as file:
        return copy_file(bytesio, file)


def tar_mkdir(file_out, name: str):
    """
    Creates an empty directory in a given tar archive on disk.

    Args:
        file_out: The output file.
        name: Name of the directory to be created.
    """
    with tarfile.TarFile("tar-archive", "a", file_out) as tfile_out:
        tarinfo = tfile_out.tarinfo()
        tarinfo.mode = 0o0755
        tarinfo.mtime = time.time()
        tarinfo.name = name
        tarinfo.tarfile = tfile_out
        tarinfo.type = tarfile.DIRTYPE
        tarinfo.uid = tarinfo.gid = 0
        tarinfo.uname = tarinfo.gname = ""
        tfile_out.addfile(tarinfo)


# TODO: What is the type of "content"?
def tar_add_file(file_out, name: str, content):
    """
    Creates a file from memory in a given tar archive on disk.

    Args:
        file_out: The output file.
        name: The name of the file to be created
        content: The corresponding file content.
    """
    bytesio = io.BytesIO(content)
    bytesio.seek(0)
    with tarfile.TarFile("tar-archive", "a", file_out) as tfile_out:
        tarinfo = tfile_out.tarinfo()
        tarinfo.mode = 0o0644
        tarinfo.mtime = time.time()
        tarinfo.name = name
        tarinfo.tarfile = tfile_out
        tarinfo.size = len(content)
        tarinfo.uid = tarinfo.gid = 0
        tarinfo.uname = tarinfo.gname = ""
        tfile_out.addfile(tarinfo, bytesio)


def tar_delete_file(file_out, name: str):
    """
    Removes a file from a given tar archive on disk.

    Args:
        file_out: The output file.
        name: The name of the file to be removed.
    """
    with tempfile.TemporaryFile() as tmp:
        with tarfile.open(fileobj=file_out) as tfile_in:
            with tarfile.open(fileobj=tmp, mode="w:") as tfile_out:
                for tarinfo in tfile_in:
                    if tarinfo.name == name:
                        continue

                    if tarinfo.type == tarfile.DIRTYPE:
                        tfile_out.addfile(tarinfo)
                    else:
                        handle = tfile_in.extractfile(tarinfo)
                        tfile_out.addfile(tarinfo, handle)
        file_out.seek(0)
        tmp.seek(0)
        shutil.copyfileobj(tmp, file_out)


def tar(file_out, name: str, file_in):
    """
    Adds a file to a given tar archive on disk to disk.

    Args:
        file_out: The output file (tar).
        name: The name of the file to be added.
        file_in: The input file.
    """
    with tarfile.TarFile("tar-archive", "a", file_out) as tfile_out:
        tarinfo = tfile_out.gettarinfo(None, name, file_in)
        tarinfo.uid = tarinfo.gid = 0
        tarinfo.uname = tarinfo.gname = ""
        tarinfo.mode = 0o0644
        tfile_out.addfile(tarinfo, file_in)


def untar(file_in, name: str, file_out):
    """
    Extracts a file from a given tar archive.

    Args:
        file_in: The input file (tar).
        name: The name of the file to be extracted.
        file_out: The output file.
    """
    result = None
    with tarfile.open(fileobj=file_in) as tfile_in:
        for tarinfo in tfile_in:
            if tarinfo.name == name:
                result = _chunk_file(tfile_in.extractfile(tarinfo), file_out)
                break
    return result


def file_exists_in_tar(file_in, name: str):
    """
    Checks if a file exists in a given tar archive.

    Args:
        file_in: The input file (tar).
        name: The name of the file for which to check existance.

    Returns:
        bool: True if the file exists, False otherwise.
    """
    with tarfile.open(fileobj=file_in) as tfile_in:
        for tarinfo in tfile_in:
            if tarinfo.name == name:
                return True
    return False


def gunzip(file_in, file_out) -> Dict:
    """
    Uncompresses a given gzip archive.

    Args:
        file_in: The input file (gz)
        file_out: The output file.

    Returns:
        dict: as defined by :func:~docker_sign_verify.Utils._chunk_file.
    """
    with gzip.GzipFile(fileobj=file_in, mode="rb") as gfile_in:
        return _chunk_file(gfile_in, file_out)



#!/usr/bin/env python

"""Stub classes for offline testing."""

from docker_sign_verify import (
    FormattedSHA256,
    ImageConfig,
    ImageName,
    ImageSource,
    Manifest,
    RegistryV2Manifest,
    Signer,
)

from .testutils import get_test_data


def _signer_for_signature(signature: str):
    """Override of docker_sign_verify.Signer::_for_signature()."""
    if "FAKE SIGNATURE" in signature:
        return FakeSigner()
    raise RuntimeError("Unsupported signature type!")


class FakeSigner(Signer):
    """Creates and verifies docker image signatures static strings."""

    def __init__(
        self,
        signature_value="-----BEGIN FAKE SIGNATURE-----\nDEFAULT FAKE SIGNATURE\n-----END FAKE SIGNATURE-----",
    ):
        self.signature_value = signature_value

    # Signer Members

    def sign(self, data: bytes) -> str:
        return self.signature_value

    def verify(self, data: bytes, signature: str):
        return {"type": "fake", "valid": True}


class FakeRegistryV2ImageSourceNoLabels(ImageSource):
    """Fake image source used to expose methods in the abstract base class."""

    def __init__(self, request, layer_exists: bool = True, **kwargs):
        super(FakeRegistryV2ImageSourceNoLabels, self).__init__(**kwargs)
        self.config = None
        self.does_layer_exists = layer_exists
        self.manifest = None
        self.request = request

    def quick_sign(self, image_name: ImageName) -> dict:
        """
        Signs a given image in an image source using a fake signer and returns the results.
        This method is a testing shortcut.

        Args:
            image_name: The name of the image to be signed.

        Returns:
            The results of the docker_sign_verify.ImageSource::_sign_image_config() method.
        """
        # pylint: disable=protected-access
        result = self._sign_image_config(FakeSigner(), image_name)

        self.config = result["image_config"]
        self.manifest.set_config_digest(
            result["image_config"].get_config_digest(),
            len(result["image_config"].get_config()),
        )
        return result

    # ImageSource Members

    def get_image_config(self, image_name: ImageName) -> ImageConfig:
        if not self.config:
            config = get_test_data(self.request, __name__, "stub_config.json")
            self.config = ImageConfig(config)
        return self.config

    def get_image_layer_to_disk(self, image_name: ImageName, layer: str, file):
        raise RuntimeError("Logic error; method should not be invoked!")

    def get_manifest(self, image_name: ImageName = None) -> Manifest:
        if not self.manifest:
            manifest = get_test_data(self.request, __name__, "stub_manifest.json")
            self.manifest = RegistryV2Manifest(manifest)
        return self.manifest

    def put_manifest(self, manifest: Manifest, image_name: ImageName = None):
        raise RuntimeError("Logic error; method should not be invoked!")

    def put_image_config(self, image_name: ImageName, image_config: ImageConfig):
        raise RuntimeError("Logic error; method should not be invoked!")

    def put_image_layer(self, image_name: ImageName, content):
        raise RuntimeError("Logic error; method should not be invoked!")

    def put_image_layer_from_disk(self, image_name: ImageName, file):
        raise RuntimeError("Logic error; method should not be invoked!")

    def layer_exists(self, image_name: ImageName, layer: FormattedSHA256) -> bool:
        return self.does_layer_exists

    def sign_image(
        self,
        signer: Signer,
        src_image_name: ImageName,
        dest_image_source,
        dest_image_name: ImageName,
    ):
        raise RuntimeError("Logic error; method should not be invoked!")

    def verify_image_integrity(self, image_name: ImageName):
        data = self._verify_image_config(image_name)

        # LGTM ...

        return {
            "compressed_layer_files": [],
            "image_config": data["image_config"],
            "manifest": data["manifest"],
            "uncompressed_layer_files": [],
        }



#!/usr/bin/env python

"""Utilities for verifying docker image integrity and signatures."""

# TODO: Revamp interfaces to conform with https://github.com/opencontainers/image-spec/blob/master/media-types.md

from .manifests import (
    ArchiveManifest,
    DeviceMapperRepositoryManifest,
    Manifest,
    RegistryV2Manifest,
)
from .signers import GPGSigner, PKISigner, Signer
from .imageconfig import ImageConfig
from .imagename import ImageName
from .imagesources import (
    ArchiveImageSource,
    DeviceMapperRepositoryImageSource,
    ImageSource,
    RegistryV2ImageSource,
)
from .utils import FormattedSHA256


__version__ = "0.5.0"



#!/usr/bin/env python

"""Classes that provide signature functionality."""

# TODO: Update pycharm version and test
# from __future__ import annotations

import abc
import base64
import logging
import io
import os
import re
import tempfile

from pathlib import Path
from typing import List

import gnupg

from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from Crypto.Signature.pkcs1_15 import PKCS115_SigScheme
from Crypto.Hash import SHA256

LOGGER = logging.getLogger(__name__)


class Signer(abc.ABC):
    """
    Abstract class to create and verify docker image signatures.
    """

    @staticmethod
    def for_signature(signature: str):
        """
        Retrieves a signer that can be used to verify a given signature value.

        Args:
            signature: The PEM encoded signature value for which to retrieve the signer.

        Returns:
            The corresponding signer.
        """
        result = None

        if "PGP SIGNATURE" in signature:
            result = GPGSigner()
        elif "PKI SIGNATURE" in signature:
            result = PKISigner()

        if not result:
            raise RuntimeError("Unsupported signature type!")

        return result

    @abc.abstractmethod
    def sign(self, data: bytes) -> str:
        """
        Signs given data.

        Args:
            data: The data to be signed.

        Returns:
            A PEM encoded signature value.
        """

    @abc.abstractmethod
    def verify(self, data: bytes, signature: bytes):
        """
        Verifies data against a given signature.

        Args:
            data: The data to be verified.
            signature: The signature against which to verify the data.

        Returns:
            Signer-specific result value.
        """


class GPGSigner(Signer):
    """
    Creates and verifies docker image signatures using GnuPG.
    """

    HOMEDIR = os.environ.get("DSV_GPG_DATASTORE", Path.home().joinpath(".gnupg"))

    def __init__(self, keyid=None, passphrase=None, homedir=HOMEDIR):
        """
        Args:
            keyid: The GPG key identifier, only required for signing.
            passphrase: The passphrase used to unlock the GPG key.
            homedir: The GPG home directory (default: ~/.gnupg).
        """
        self.keyid = keyid
        self.passphrase = passphrase

        LOGGER.debug("Using trust store: %s", homedir)
        self.gpg = gnupg.GPG(homedir=homedir)

    def _debug_init_store(self, name="DSV Test Key", email="test@key.com"):
        """
        Initializes a new GPG keystore for testing purposes.

        Args:
            name: GPG identity name used to create the new key.
            email: GPG identity email used to create the new key.

        Returns:
            The GPG key identifier of the newly created key.
        """
        input_data = self.gpg.gen_key_input(
            name_email=name, name_real=email, passphrase=self.passphrase
        )

        result = self.gpg.gen_key(input_data)
        self.keyid = str(result)

        return result

    # Signer Members

    def sign(self, data: bytes) -> str:
        if not self.keyid:
            LOGGER.warning("Signing using implicit / default keyid!")
            # raise RuntimeError("Cannot sign without keyid!")

        kwargs = {}
        if self.keyid:
            kwargs = {"default_key": self.keyid}

        result = self.gpg.sign(
            data, clearsign=False, detach=True, passphrase=self.passphrase, **kwargs
        )

        return str(result).rstrip()

    def verify(self, data: bytes, signature: str):
        # Note: gnupg.py:verify_file() forces sig_file to be on disk, as the
        #       underlying gpg utility does the same =(
        with tempfile.NamedTemporaryFile() as tmpfile:
            tmpfile.write(signature.encode("utf-8"))
            tmpfile.flush()
            os.fsync(tmpfile.fileno())
            return self.gpg.verify_file(io.BytesIO(data), tmpfile.name)


# https://pysheeet.readthedocs.io/en/latest/notes/python-security.html
class PKISigner(Signer):
    """
    Creates and verifies docker image signatures using OpenSSL.
    """

    KEYPAIR = os.environ.get("DSV_PKI_DATASTORE", Path.home().joinpath("dsv.pem"))

    TAG_START = "-----BEGIN PKI SIGNATURE-----"
    TAG_END = "-----END PKI SIGNATURE-----"

    def __init__(self, keypair_path=KEYPAIR, passphrase=None):
        """
        Reference: https://www.digicert.com/ssl-support/pem-ssl-creation.htm
        Args:
            keypair_path: Path to a file containing the entire trust chain and private key, PEM encoded.
            passphrase: The passphrase used to unlock the private key.
        """
        self.keypair_entries = None
        self.keypair_path = keypair_path
        self.passphrase = passphrase
        self.private_signer = None
        self.public_signer = None

        LOGGER.debug("Using keypair: %s", self.keypair_path)

    def _debug_init_keypair(self, bits=2048):
        """
        Initializes a new PKI certificate pair.

        Args:
            bits: Entropy to use when generating the private key.
        """
        key = RSA.generate(bits)
        private_key = key.export_key(passphrase=self.passphrase)
        public_key = key.publickey().export_key()

        with open(self.keypair_path, "wb") as file:
            file.write(public_key)
            file.write(os.linesep.encode("utf-8"))
            file.write(private_key)
            file.flush()

    def get_keypair(self) -> List:
        """
        Retrieves the keypair entries from disk.

        Returns:
            The list of PEM encoded entries.
        """
        if self.keypair_entries is None:
            with open(self.keypair_path, "r") as file:
                content = file.read()
            self.keypair_entries = re.findall(
                r"(-{5}BEGIN([^-]+)-{5}.+-{5}END\2-{5})", content, re.DOTALL
            )
            self.keypair_entries = [x[0] for x in self.keypair_entries]
        return self.keypair_entries

    def get_private_signer(self) -> PKCS115_SigScheme:
        """
        Retrieves a private signer, initialized from the first PEM encoded private key entry.

        Returns:
            A private signer initialized from the first PEM encoded private key entry.
        """
        if not self.private_signer:
            for entry in self.get_keypair():
                if "RSA PRIVATE KEY" in entry:
                    private_rsa_key = RSA.import_key(entry, self.passphrase)
                    self.private_signer = PKCS1_v1_5.new(private_rsa_key)
        return self.private_signer

    def get_public_signer(self) -> PKCS115_SigScheme:
        """
        Retrieves a public signer, initialized from the first PEM encoded public key entry.

        Returns:
            A public signer, initialized from the first PEM encoded public key entry.
        """
        if not self.public_signer:
            for entry in self.get_keypair():
                if "CERTIFICATE" in entry or "PUBLIC KEY" in entry:
                    # TODO: Collect x509 certificate details
                    # TODO: Is it possible to use OS-level cacerts instead of providing the public key?
                    public_rsa_key = RSA.import_key(entry)
                    self.public_signer = PKCS1_v1_5.new(public_rsa_key)
        return self.public_signer

    # Signer Members

    def sign(self, data: bytes) -> str:
        digest = SHA256.new(data)
        raw_signature = self.get_private_signer().sign(digest)

        return "{0}\n\n{1}\n{2}".format(
            PKISigner.TAG_START,
            base64.b64encode(raw_signature).decode(),
            PKISigner.TAG_END,
        )

    def verify(self, data: bytes, signature: str):
        # if not self.public_key_path:
        #    raise RuntimeError("Cannot verify without public key!")
        buffer = base64.b64decode(signature.split()[3])

        digest = SHA256.new(data)
        # pylint: disable=not-callable
        valid = False
        try:
            self.get_public_signer().verify(digest, buffer)
            valid = True
        except ValueError:
            ...

        return {"keypair_path": self.keypair_path, "type": "pki", "valid": valid}



#!/usr/bin/env python

"""Utility classes."""


from docker_sign_verify import ImageName


# pylint: disable=unused-argument
def to_image_name(context, param, value: str) -> ImageName:
    """Converts an docker image name to an ImageName."""
    if isinstance(value, str):
        result = ImageName.parse(value)
    else:
        result = [ImageName.parse(v) for v in value]
    return result


# pylint: disable=too-few-public-methods
class HiddenPassword:
    """Helper class to mask password input."""

    def __init__(self, password: str = ""):
        """
        Args:
            password: The password.
        """
        self.password = password

    def __str__(self):
        return "*" * len(self.password)



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""ImageName tests."""

import pytest

from docker_sign_verify import ImageName


@pytest.fixture()
def url():
    """Provides a fake URL with endpoint, port, namespace, image, and tag segments."""
    return "endpoint:port/namespace/image:tag"


@pytest.fixture()
def image_name(url: str) -> ImageName:
    """Provides ImageName instance for the fake URL."""
    return ImageName.parse(url)


def test_init():
    """Test that image name can be instantiated."""
    assert ImageName("endpoint:port", "namespace/image", "tag")


def test_str():
    """Test __str__ pass-through for different variants."""
    for endpoint in ["endpoint.io", "endpoint:port", None]:
        for image in ["image", "namespace/image"]:
            for tag in ["tag", None]:
                image_name = ImageName(endpoint, image, tag)
                string = str(image_name)
                assert image in string
                if endpoint:
                    assert endpoint in string
                if tag:
                    assert tag in string
                assert "None" not in string


def test_parse_string():
    """Test string parsing for complex image names."""
    for endpoint in ["endpoint.io", "endpoint:port", None]:
        for image in ["image", "namespace/image"]:
            for tag in ["tag", None]:

                # Construct a complex string ...
                string = image
                if tag:
                    string = "{0}:{1}".format(string, tag)
                if endpoint:
                    string = "{0}/{1}".format(endpoint, string)

                # Verify the complex string was parsed correctly ...
                # pylint: disable=protected-access
                result = ImageName._parse_string(string)
                assert result["endpoint"] == endpoint
                assert result["image"] == image
                assert result["tag"] == tag


def test_parse():
    """Test initialization via parsed strings."""
    (endpoint, image, tag) = ["address:port", "namespace/image", "tag"]
    image_name = ImageName.parse("{0}/{1}:{2}".format(endpoint, image, tag))
    assert image_name.endpoint == endpoint
    assert image_name.image == image
    assert image_name.tag == tag


def test_endpoint(image_name: ImageName):
    """Tests endpoint retrieval."""
    assert image_name.endpoint == "endpoint:port"


def test_image(image_name: ImageName):
    """Tests endpoint image."""
    assert image_name.image == "namespace/image"


def test_tag(image_name: ImageName):
    """Tests endpoint tag."""
    assert image_name.tag == "tag"


# TODO: Test resolve methods ...



#!/usr/bin/env python

"""Class that provides parsing and formatting of docker image names."""

import os

from typing import Dict, Union


class ImageName:
    """
    Docker image name abstraction.
    """

    DEFAULT_REGISTRY_ENDPOINT = os.environ.get(
        "DSV_DEFAULT_REGISTRY", "index.docker.io"
    )

    DEFAULT_REGISTRY_NAMESPACE = os.environ.get("DSV_DEFAULT_NAMESPACE", "library")

    DEFAULT_REGISTRY_TAG = os.environ.get("DSV_DEFAULT_TAG", "latest")

    def __init__(self, endpoint: Union[str, None], image: str, tag: str = None):
        """
        Args:
            endpoint: Endpoint address for fully qualified image names.
            image: Name of the image, optionally including a namespace.
            tag: Optional tag name.
        """
        self.endpoint = endpoint
        self.image = image
        self.tag = tag

    def __str__(self):
        result = self.image
        if self.tag:
            result = "{0}:{1}".format(result, self.tag)
        if self.endpoint:
            result = "{0}/{1}".format(self.endpoint, result)

        return result

    @staticmethod
    def _parse_string(string: str) -> Dict:
        """
        Parses the endpoint, image, and tag from a given string.

        Args:
            string: The string to be parsed.

        Returns:
            dict:
                endpoint: The registry endpoint; address with optional port.
                image: The name of the image; the image name and optional namespace.
                tag: The tag name.
        """
        result = {"endpoint": None, "image": None, "tag": None}

        segments = string.split("/")

        parts = segments[-1].split(":")
        result["image"] = parts[0]
        if len(parts) == 2:
            result["tag"] = parts[1]

        if len(segments) == 1:
            # image[:tag]
            pass
        elif len(segments) == 2:
            # host[:port]/image[:tag] OR namespace/image[:tag]

            # Note: https://docs.docker.com/engine/reference/commandline/tag/
            #
            #       An image name is made up of slash-separated name components, optionally prefixed by a registry
            #       hostname. The hostname must comply with standard DNS rules, but may not contain underscores. If a
            #       hostname is present, it may optionally be followed by a port number in the format :8080. ... Name
            #       components may contain lowercase letter, digits and separators. A separator is defined as a period,
            #       one or two underscores, or one or more dashes. A name component may not start or end with a
            #       separator. A tag name ... may contain lowercase and uppercase letters, digits, underscores, periods
            #       and dashes. A tag name may not start with a period or a dash ... .

            # Assumption: That endpoint addresses will contain at least one '.' (period) character, and by convention
            #             images namespaces will not.
            if ":" not in segments[0] and "." not in segments[0]:
                result["image"] = "{0}/{1}".format(segments[0], result["image"])
            else:
                result["endpoint"] = segments[0]
        elif len(segments) == 3:
            # host[:port]/namespace/image[:tag]
            result["endpoint"] = segments[0]
            result["image"] = "{0}/{1}".format(segments[1], result["image"])
        else:
            raise RuntimeError("Unable to parse string: {0}".format(string))

        return result

    @staticmethod
    def parse(image_name: str):
        """
        Initializes an ImageName from a given image name string.

        Args:
            image_name: String containing the image name to be parsed.

        Returns:
            The newly initialized object.
        """
        parsed = ImageName._parse_string(image_name)
        return ImageName(parsed["endpoint"], parsed["image"], parsed["tag"])

    def resolve_endpoint(self) -> str:
        """
        Resolves the registry endpoint.

        Returns:
            The explicit registry endpoint.
        """
        if self.endpoint:
            return self.endpoint

        return ImageName.DEFAULT_REGISTRY_ENDPOINT

    def resolve_image(self) -> str:
        """
        Resolves the name of the image.

        Returns:
            The explicit name of the image, with namespace.
        """

        segments = self.image.split("/")
        if len(segments) < 2:
            return "{0}/{1}".format(ImageName.DEFAULT_REGISTRY_NAMESPACE, self.image)

        return self.image

    def resolve_tag(self) -> str:
        """
        Retrieves resolves the tag name.

        Returns:
            The explicit tag name
        """
        if self.tag:
            return self.tag

        return ImageName.DEFAULT_REGISTRY_TAG



#!/usr/bin/env python

"""Docker sign command line interface."""

import logging
import os

import click
import urllib3

from docker_sign_verify import (
    ArchiveImageSource,
    DeviceMapperRepositoryImageSource,
    ImageName,
    RegistryV2ImageSource,
)

from .common import logging_options, set_log_levels, version
from .utils import to_image_name, HiddenPassword

LOGGER = logging.getLogger(__name__)

# Bug Fix: There isn't anything we can do about mis-configured remote certificates ...
urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)


def sign_options(function):
    """Common signature creation options."""
    function = click.option(
        "-k",
        "--keyid",
        help="Signing key identifier.",
        required=True,
        envvar="DSV_KEYID",
    )(function)
    function = click.option(
        "-p",
        "--keypass",
        default=lambda: HiddenPassword(os.environ.get("DSV_KEYPASS", "")),
        help="Signing key passphrase.",
        hide_input=True,
        prompt=True,
    )(function)
    function = click.option(
        "-t",
        "--sigtype",
        default="gpg",
        envvar="DSV_KEYTYPE",
        help="Signature type.",
        show_default=True,
        type=click.Choice(["gpg", "pki"]),
    )(function)

    # Note: Order of argument is reversed, due to nesting ...
    # TODO: Should we support dissimilar src and dest image source types in the CLI?
    function = click.argument("dest_image_name", callback=to_image_name, required=True)(
        function
    )
    function = click.argument("src_image_name", callback=to_image_name, required=True)(
        function
    )
    return function


def sign(context):
    """Signs an image."""

    if context.obj["sigtype"] == "gpg" and (
        "." in context.obj["keyid"] or "/" in context.obj["keyid"]
    ):
        LOGGER.warning("Key identifier looks like path, but signature type is GPG!")

    signer_module = __import__("docker_sign_verify.signers")
    signer_class = getattr(
        signer_module, "{0}Signer".format(context.obj["sigtype"].upper())
    )
    signer = signer_class(context.obj["keyid"], context.obj["keypass"])

    data = context.obj["imagesource"].sign_image(
        signer,
        context.obj["src_image_name"],
        context.obj["imagesource"],
        context.obj["dest_image_name"],
    )
    LOGGER.info(
        "Created new image: %s (%s)",
        context.obj["dest_image_name"],
        data["image_config"].get_config_digest(),
    )


@click.group()
@click.option(
    "--dry-run", help="Do not write to destination image sources.", is_flag=True
)
@logging_options
@click.pass_context
def cli(context, dry_run: False, verbosity: int = 2):
    """Creates and embeds signatures into docker images."""

    set_log_levels(verbosity)

    context.obj = {"dry_run": dry_run}


@cli.command()
@click.option(
    "-a",
    "--archive",
    help="Uncompressed docker image archive.",
    required=True,
    type=click.Path(dir_okay=False, resolve_path=True),
)
@sign_options
@click.pass_context
# pylint: disable=redefined-outer-name,too-many-arguments
def archive(
    context,
    keyid: str,
    keypass: str,
    sigtype: str,
    src_image_name: ImageName,
    dest_image_name: ImageName,
    archive: str,
):
    """Operates on docker-save produced archives."""

    context.obj["keyid"] = keyid
    context.obj["keypass"] = keypass
    context.obj["sigtype"] = sigtype
    context.obj["src_image_name"] = src_image_name
    context.obj["dest_image_name"] = dest_image_name
    context.obj["imagesource"] = ArchiveImageSource(
        archive=archive, dry_run=context.obj["dry_run"]
    )
    sign(context)


@cli.command()
@sign_options
@click.pass_context
# pylint: disable=too-many-arguments
def registry(
    context,
    keyid: str,
    keypass: str,
    sigtype: str,
    src_image_name: ImageName,
    dest_image_name: ImageName,
):
    """Operates on docker registries (v2)."""

    context.obj["keyid"] = keyid
    context.obj["keypass"] = keypass
    context.obj["sigtype"] = sigtype
    context.obj["src_image_name"] = src_image_name
    context.obj["dest_image_name"] = dest_image_name
    context.obj["imagesource"] = RegistryV2ImageSource(dry_run=context.obj["dry_run"])
    sign(context)


@cli.command()
@sign_options
@click.pass_context
# pylint: disable=too-many-arguments
def repository(
    context,
    keyid: str,
    keypass: str,
    sigtype: str,
    src_image_name: ImageName,
    dest_image_name: ImageName,
):
    """Operates on docker repositories."""

    context.obj["keyid"] = keyid
    context.obj["keypass"] = keypass
    context.obj["sigtype"] = sigtype
    context.obj["src_image_name"] = src_image_name
    context.obj["dest_image_name"] = dest_image_name
    context.obj["imagesource"] = DeviceMapperRepositoryImageSource(
        dry_run=context.obj["dry_run"]
    )
    sign(context)


cli.add_command(version)

if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()



#!/usr/bin/env python

"""Common CLI commands."""

import logging
import sys

import click

from docker_sign_verify import __version__


def logging_options(function):
    """Common logging options."""

    function = click.option(
        "-s", "--silent", "verbosity", flag_value=0, help="Suppress all output."
    )(function)
    function = click.option(
        "-q",
        "--quiet",
        "verbosity",
        flag_value=1,
        help="Restrict output to warnings and errors.",
    )(function)
    function = click.option(
        "-d",
        "--debug",
        "-v",
        "--verbose",
        "verbosity",
        flag_value=3,
        help="Show debug logging.",
    )(function)
    function = click.option(
        "-vv", "--very-verbose", "verbosity", flag_value=4, help="Enable all logging."
    )(function)

    return function


def set_log_levels(verbosity):
    """
    Assigns the logging levels in a consistent way.

    Args:
        verbosity: The logging verbosity level from  0 (least verbose) to 4 (most verbose).
    """
    levels = {1: logging.WARNING, 2: logging.INFO, 3: logging.DEBUG, 4: logging.NOTSET}

    if verbosity is None:
        verbosity = 2

    if verbosity < 3:
        logging.basicConfig(
            stream=sys.stdout, level=levels[verbosity], format="%(message)s"
        )
        logging.getLogger("gnupg").setLevel(logging.FATAL)
    else:
        logging.basicConfig(stream=sys.stdout, level=levels[verbosity])

    if verbosity == 3:
        logging.getLogger("gnupg").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)


@click.command()
def version():
    """Displays the utility version."""

    print(__version__)



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""Manifest tests."""

import json
import pytest

from docker_sign_verify import (
    ArchiveManifest,
    DeviceMapperRepositoryManifest,
    FormattedSHA256,
    ImageName,
    RegistryV2Manifest,
)

from .testutils import get_test_data


@pytest.fixture()
def formattedsha256() -> FormattedSHA256:
    """Provides a FormattedSHA256 instance with a distinct digest value."""
    return FormattedSHA256(
        "ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789"
    )


@pytest.fixture()
def sha256_archive_layer(formattedsha256: FormattedSHA256) -> str:
    """Provides the archive layer identifier for the distinct digest value."""
    return "{0}/layer.tar".format(formattedsha256.sha256)


@pytest.fixture
def manifest_archive(request) -> bytes:
    """Provides a sample archive manifest."""
    return get_test_data(request, __name__, "manifest_archive.json")


@pytest.fixture
def manifest_registry(request) -> bytes:
    """Provides a sample registry manifest."""
    return get_test_data(request, __name__, "manifest_registry.json")


@pytest.fixture
def manifest_repository(request) -> bytes:
    """Provides a sample repository manifest."""
    return get_test_data(request, __name__, "manifest_repository.json")


@pytest.fixture
def archive_manifest(manifest_archive: bytes) -> ArchiveManifest:
    """Provides an ArchiveManifest instance for the sample archive manifest."""
    # Do not use caching; get a new instance for each test
    return ArchiveManifest(manifest_archive)


@pytest.fixture
def devicemapper_repository_manifest(
    manifest_repository: bytes
) -> DeviceMapperRepositoryManifest:
    """Provides a DeviceMapperRepositoryManifest instance for the sample repository manifest."""
    # Do not use caching; get a new instance for each test
    return DeviceMapperRepositoryManifest(manifest_repository)


@pytest.fixture
def registry_v2_manifest(manifest_registry: bytes) -> RegistryV2Manifest:
    """Provides a RegistryV2Manifest instance for the sample registry manifest."""
    # Do not use caching; get a new instance for each test
    return RegistryV2Manifest(manifest_registry)


def test_init(
    archive_manifest: ArchiveManifest,
    devicemapper_repository_manifest: DeviceMapperRepositoryManifest,
    registry_v2_manifest: RegistryV2Manifest,
):
    """Test that all manifest types can be instantiated."""
    assert archive_manifest
    assert devicemapper_repository_manifest
    assert registry_v2_manifest


def test_str(
    archive_manifest: ArchiveManifest,
    devicemapper_repository_manifest: DeviceMapperRepositoryManifest,
    registry_v2_manifest: RegistryV2Manifest,
    manifest_archive: bytes,
    manifest_registry: bytes,
    manifest_repository: bytes,
):
    """Test __str__ pass-through for all manifest types (with encoding)."""

    # Note: Due to whitespace differences, compare the parsed string form.
    assert json.loads(str(archive_manifest)) == json.loads(manifest_archive)
    assert json.loads(str(devicemapper_repository_manifest)) == json.loads(
        manifest_repository
    )
    assert json.loads(str(registry_v2_manifest)) == json.loads(manifest_registry)


def test_archive__layer_to_hash(
    archive_manifest: ArchiveManifest,
    sha256_archive_layer: str,
    formattedsha256: FormattedSHA256,
):
    """Test layer to hash conversion."""
    assert archive_manifest.layer_to_digest(sha256_archive_layer) == formattedsha256


def test_archive__hash_to_layer(
    archive_manifest: ArchiveManifest,
    sha256_archive_layer: str,
    formattedsha256: FormattedSHA256,
):
    """Test hash to layer conversion."""
    assert archive_manifest.digest_to_layer(formattedsha256) == sha256_archive_layer


@pytest.mark.parametrize(
    "img_name,config_value",
    [
        (
            "a51f3f9281a1a3d89dce25fec8acffbe9f59ddb67d98e04245c4c886e32d3782",
            "a51f3f9281a1a3d89dce25fec8acffbe9f59ddb67d98e04245c4c886e32d3782.json",
        ),
        (
            "base:7.2",
            "adecf4209bb9dd67d96393774cbd7f8bd2bad3596da42cde33daa0c41b14ac62.json",
        ),
        (
            "adecf4209bb9dd67d96393774cbd7f8bd2bad3596da42cde33daa0c41b14ac62",
            "adecf4209bb9dd67d96393774cbd7f8bd2bad3596da42cde33daa0c41b14ac62.json",
        ),
    ],
)
def test_archive__get_config(
    archive_manifest: ArchiveManifest, img_name: str, config_value: str
):
    """Test configuration retrieval."""
    image_name = ImageName.parse(img_name)
    assert archive_manifest.get_config(image_name)["Config"] == config_value


@pytest.mark.parametrize("repotag", ["REPOTAG", None])
def test_archive_append_config(
    archive_manifest: ArchiveManifest,
    sha256_archive_layer: str,
    formattedsha256: FormattedSHA256,
    repotag,
):
    """Test appending configurations."""
    layers = [formattedsha256]
    archive_manifest.append_config(formattedsha256, layers, repotag)
    image_name = ImageName(None, formattedsha256.sha256)
    config = archive_manifest.get_config(image_name)
    assert config["Config"] == "{0}.json".format(formattedsha256.sha256)
    assert config["Layers"] == [sha256_archive_layer]
    assert config.get("RepoTags", None) == repotag


def test_archive_get_config_digest(archive_manifest: ArchiveManifest):
    """Test configuration digest retrieval."""
    formattedsha256 = FormattedSHA256(
        "a51f3f9281a1a3d89dce25fec8acffbe9f59ddb67d98e04245c4c886e32d3782"
    )
    image_name = ImageName.parse(formattedsha256.sha256)
    assert archive_manifest.get_config_digest(image_name) == formattedsha256


def test_archive_get_layers(archive_manifest: ArchiveManifest):
    """Test manifest layer retrieval."""
    sha256 = "a51f3f9281a1a3d89dce25fec8acffbe9f59ddb67d98e04245c4c886e32d3782"
    image_name = ImageName.parse(sha256)
    assert archive_manifest.get_layers(image_name) == [
        "sha256:137120c8596a15ab42c39c0c5cf83ef864b6b65b5516887c895915e87292bd07",
        "sha256:755520f73bc74ae73b12f53229e401e8d4c584b74f5704d2d36ba7c45e2657cf",
        "sha256:13fb089903a5e0e9b00d78ba48496da528ce8d81e08a1042ebeced8c35d714cb",
        "sha256:f86d68f70ca006025a7f7013f69898f78d1d9272c4d3909e3ec4c7f9958da20e",
        "sha256:7b4a4edd704242cec1710679a088be8aabff25c3a79f4eecbe8d11d57c53a20b",
        "sha256:ef4724d42630f3022ef67c3f6749e85a13e81b8efcf98fbd517476499f10e5ab",
    ]


def test_registry_override_config(
    registry_v2_manifest: RegistryV2Manifest, formattedsha256: FormattedSHA256
):
    """Test overriding manifest values."""
    size = 1234
    registry_v2_manifest.set_config_digest(formattedsha256, size)
    assert registry_v2_manifest.get_config_digest() == formattedsha256
    assert registry_v2_manifest.json["config"]["size"] == size


def test_registry_get_config_digest(registry_v2_manifest: RegistryV2Manifest):
    """Test image configuration digest retrieval."""
    formattedsha256 = FormattedSHA256(
        "8f1196ff19e7b5c5861de192ae77e8d7a692fcbca2dd3174d324980f72ab49bf"
    )
    assert registry_v2_manifest.get_config_digest() == formattedsha256
    assert (
        registry_v2_manifest.get_config_digest(ImageName.parse("ignored"))
        == formattedsha256
    )


def test_registry_get_layers(registry_v2_manifest: RegistryV2Manifest):
    """Test manifest layer retrieval."""
    layers = [
        "sha256:6c8c72249e560701aa1da4cd40192274a8c0419ddb8e4a553aa02b5a1acdb863",
        "sha256:1403b179e2c9df4f57e9ea94e32882739c6b3d75ed756d4e67fcc424288c29cc",
    ]
    assert registry_v2_manifest.get_layers() == layers
    assert registry_v2_manifest.get_layers(ImageName.parse("ignored")) == layers


@pytest.mark.parametrize(
    "img_name", ["endpoint:port/namespace/image:tag", "rancher/swarmkit:v1.13.0-beta.1"]
)
def test_repository_override_config(
    devicemapper_repository_manifest: DeviceMapperRepositoryManifest,
    formattedsha256: FormattedSHA256,
    img_name: str,
):
    """Test overriding manifest values."""
    image_name = ImageName.parse(img_name)
    devicemapper_repository_manifest.override_config(formattedsha256, image_name)
    assert (
        devicemapper_repository_manifest.get_config_digest(image_name)
        == formattedsha256
    )


@pytest.mark.parametrize(
    "img_name,config_digest",
    [
        (
            "busybox:latest",
            "sha256:6ad733544a6317992a6fac4eb19fe1df577d4dec7529efec28a5bd0edad0fd30",
        )
    ],
)
def test_repository_get_config_digest(
    devicemapper_repository_manifest: DeviceMapperRepositoryManifest,
    img_name: str,
    config_digest: str,
):
    """Test image configuration digest retrieval."""
    image_name = ImageName.parse(img_name)
    assert (
        devicemapper_repository_manifest.get_config_digest(image_name) == config_digest
    )


# def test_repository_get_layers(devicemapper_repository_manifest: DeviceMapperRepositoryManifest):
#    """Test manifest layer retrieval."""
#    with pytest.raises(NotImplementedError):
#        devicemapper_repository_manifest.get_layers()



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""RegistryV2ImageSource tests."""

import base64
import copy
import tempfile

import pytest

from docker_sign_verify import (
    ImageConfig,
    ImageName,
    RegistryV2ImageSource,
    RegistryV2Manifest,
)
from docker_sign_verify.utils import FormattedSHA256

from .stubs import FakeSigner
from .testutils import get_test_data_path

KNOWN_GOOD_IMAGE_LIST = ["busybox:1.30.1", "library/python:3.7.2-slim-stretch"]


@pytest.fixture
def registry_v2_image_source() -> RegistryV2ImageSource:
    """Provides a RegistryV2ImageSource instance."""
    # Do not use caching; get a new instance for each test
    return RegistryV2ImageSource(dry_run=True)


def test_init(registry_v2_image_source: RegistryV2ImageSource):
    """Test that the image source can be instantiated."""
    assert registry_v2_image_source


@pytest.mark.parametrize(
    "endpoint,expected_username,expected_password",
    [
        ("endpoint:port", "username", "password"),
        ("endpoint2:port2", "username2", "password2"),
    ],
)
def test__get_credentials(
    request,
    registry_v2_image_source: RegistryV2ImageSource,
    endpoint: str,
    expected_username: str,
    expected_password: str,
):
    """Test credentials retrieval."""
    registry_v2_image_source.credentials_store = get_test_data_path(
        request, "credentials_store.json"
    )
    # pylint: disable=protected-access
    credentials = registry_v2_image_source._get_credentials(endpoint)
    assert credentials

    decoded = base64.decodebytes(credentials.encode("utf-8")).decode("utf-8")
    assert decoded

    actual_username, actual_password = decoded.split(":")
    assert actual_username == expected_username
    assert actual_password == expected_password


@pytest.mark.parametrize(
    "image,expected_credentials",
    [
        ("endpoint:port/image", "dXNlcm5hbWU6cGFzc3dvcmQ="),
        ("endpoint2:port2/image", "dXNlcm5hbWUyOnBhc3N3b3JkMg=="),
    ],
)
def test__get_request_headers(
    request,
    registry_v2_image_source: RegistryV2ImageSource,
    image: str,
    expected_credentials: str,
):
    """Test request headers retrieval."""
    registry_v2_image_source.credentials_store = get_test_data_path(
        request, "credentials_store.json"
    )
    image_name = ImageName.parse(image)
    # pylint: disable=protected-access
    headers = registry_v2_image_source._get_request_headers(image_name)
    assert headers
    assert expected_credentials in headers["Authorization"]


@pytest.mark.parametrize("image", KNOWN_GOOD_IMAGE_LIST)
def test_get_iamge_config(registry_v2_image_source: RegistryV2ImageSource, image: str):
    """Test image configuration retrieval."""
    image_name = ImageName.parse(image)
    config = registry_v2_image_source.get_image_config(image_name)

    assert config
    assert isinstance(config, ImageConfig)


@pytest.mark.parametrize("image", KNOWN_GOOD_IMAGE_LIST)
def test_get_image_layer_to_disk(
    registry_v2_image_source: RegistryV2ImageSource, image: str
):
    """Test layer retrieval to disk."""
    image_name = ImageName.parse(image)
    config_digest = registry_v2_image_source.get_manifest(
        image_name
    ).get_config_digest()
    temp = tempfile.NamedTemporaryFile()
    result = registry_v2_image_source.get_image_layer_to_disk(
        image_name, config_digest, temp
    )
    assert result["digest"] == config_digest


@pytest.mark.parametrize("image", KNOWN_GOOD_IMAGE_LIST)
def test_get_manifest(registry_v2_image_source: RegistryV2ImageSource, image: str):
    """Test manifest retrieval."""
    image_name = ImageName.parse(image)
    manifest = registry_v2_image_source.get_manifest(image_name)

    assert manifest
    assert isinstance(manifest, RegistryV2Manifest)


@pytest.mark.parametrize("image", KNOWN_GOOD_IMAGE_LIST)
def test_layer_exists(registry_v2_image_source: RegistryV2ImageSource, image: str):
    """Test layer existence."""
    image_name = ImageName.parse(image)
    layer = registry_v2_image_source.get_manifest(image_name).get_layers()[-1]
    assert registry_v2_image_source.layer_exists(image_name, layer)
    assert not registry_v2_image_source.layer_exists(
        image_name, FormattedSHA256("0" * 64)
    )


@pytest.mark.parametrize("image", KNOWN_GOOD_IMAGE_LIST)
def test_sign_image_same_image_source(
    registry_v2_image_source: RegistryV2ImageSource, image: str
):
    """Test image signing."""
    src_image_name = ImageName.parse(image)
    dest_image_name = copy.deepcopy(src_image_name)
    dest_image_name.tag = "{0}_signed".format(dest_image_name.tag)

    def assertions(result: dict):
        assert result

        image_config = result["image_config"]
        assert image_config
        assert "FAKE SIGNATURE" in str(image_config)

        signature_value = result["signature_value"]
        assert signature_value
        assert "FAKE SIGNATURE" in signature_value

        verify_image_data = result["verify_image_data"]
        assert verify_image_data
        assert image_config == verify_image_data["image_config"]

        manifest = verify_image_data["manifest"]
        assert manifest

        manifest_signed = result["manifest_signed"]
        assert manifest_signed
        assert manifest_signed.get_config_digest() == image_config.get_config_digest()
        assert len(manifest_signed.get_layers()) == len(image_config.get_image_layers())

    # 1. Single signature
    assertions(
        registry_v2_image_source.sign_image(
            FakeSigner(), src_image_name, registry_v2_image_source, dest_image_name
        )
    )

    # TODO: Test signing image twice (with same key, with different keys ...)
    #       Can we do this here (using dockerhub), or do we need to do this in test_imageconfig.py???


# TODO: test_sign_image_different_image_source


@pytest.mark.parametrize("image", KNOWN_GOOD_IMAGE_LIST)
def test_verify_image_integrity(
    registry_v2_image_source: RegistryV2ImageSource, image: str
):
    """Test image unsigning."""
    image_name = ImageName.parse(image)

    def assertions(result: dict):
        assert result

        image_config = result["image_config"]
        assert image_config

        manifest = result["manifest"]
        assert manifest

        assert len(result["compressed_layer_files"]) == len(
            result["uncompressed_layer_files"]
        )

    # 1. Unsigned
    assertions(registry_v2_image_source.verify_image_integrity(image_name))

    # TODO: Test integrity on a signed image ...
    #       Can we do this here (using dockerhub), or do we need to do this in test_imageconfig.py???



#!/usr/bin/env python

"""
Abstraction of a docker image configuration, as defined in:

https://github.com/moby/moby/blob/master/image/spec/v1.md
(eventually, https://github.com/opencontainers/image-spec/blob/master/config.md)
"""

import copy
import json
import logging

from typing import Dict, List

import canonicaljson

from .signers import Signer
from .utils import formatted_digest, must_be_equal, FormattedSHA256

LOGGER = logging.getLogger(__name__)


class ImageConfig:
    """
    Docker image configuration.
    """

    # Label containing a string value (escaped json)
    SIGNATURES_LABEL = "signatures"

    # The normalized empty signature value
    DEFAULT_SIGNATURES_VALUE = "[]"

    def __init__(self, config: bytes):
        """
        Args:
            config: The raw image configuration.
        """
        self.config = self.config_json = None
        self._set_config(config)

    def __str__(self):
        return self.get_config().decode("utf-8")

    @staticmethod
    def _get_labels(config_json) -> Dict:
        """
        Retrieves the "Labels" dictionary from the given image configuration.

        Args:
            config_json: The image configuration from which to retrieve the dictionary.

        Returns:
            Dict: The corresponding dictionary, or an empty dictionary if NoneType.
        """

        # Note: We need to handle both key cases, as Red Hat does not conform to the standard.
        try:
            config = config_json["Config"]
        except KeyError:
            config = config_json["config"]

        if config is None:
            raise RuntimeError(
                "Unable to locate [Cc]onfig key within image configuration!"
            )

        try:
            labels = config["Labels"]
        except KeyError:
            labels = None
            LOGGER.debug("Non-conformant image configuration; 'Labels' missing!")

        if labels is None:
            labels = config["Labels"] = {}

        return labels

    @staticmethod
    def _normalize(config_json):
        """
        Normalizes a given image configuration so that it contains, at a minimum, the signatures label.

        Args:
            config_json: The image configuration to be normalized.

        Returns:
            The normalized image configuration.
        """
        labels = ImageConfig._get_labels(config_json)
        signatures = labels.get(
            ImageConfig.SIGNATURES_LABEL, ImageConfig.DEFAULT_SIGNATURES_VALUE
        )
        labels[ImageConfig.SIGNATURES_LABEL] = signatures

        return config_json

    def _set_config(self, config: bytes):
        """
        Assigns the raw image configuration and updates the internal JSON object.

        Args:
            config: The raw image configuration.
        """
        self.config = config
        # Note: Do not normalize here, or integrity verification of unsigned images will fail.
        self.config_json = json.loads(self.config)

    def _set_config_json(self, config_json):
        """
        Assigns the internal JSON object and updates the raw image configuration.

        Args:
            config_json: The internal JSON object.
        """
        self.config_json = config_json
        self.config = json.dumps(self.config_json).encode("utf-8")

    def get_config(self) -> bytes:
        """
        Retrieves the raw image configuration.

        Returns:
            The raw image configuration.
        """
        return self.config

    def get_config_canonical(self):
        """
        Retrieves the image configuration in canonical JSON form.

        Returns:
            The image configuration in canonical JSON form.
        """
        config_json = copy.deepcopy(self.config_json)
        config_json = ImageConfig._normalize(config_json)
        return canonicaljson.encode_canonical_json(config_json)

    def get_config_digest(self) -> FormattedSHA256:
        """
        Retrieves the SHA256 digest value of the raw image configuration.

        Returns:
            The SHA256 digest value of the raw image configuration.
        """
        return formatted_digest(self.get_config())

    def get_config_digest_canonical(self) -> FormattedSHA256:
        """
        Retrieves the SHA256 digest value of the image configuration in canonical JSON form.

        Returns:
            The SHA256 digest value of the image configuration in canonical JSON form.
        """
        return formatted_digest(self.get_config_canonical())

    def get_image_layers(self) -> List:
        """
        Retrieves the listing of image layer identifiers.

        Returns:
            The listing of image layer identifiers.
        """
        diff_ids = self.config_json["rootfs"]["diff_ids"]
        return [FormattedSHA256.parse(x) for x in diff_ids]

    def clear_signature_list(self):
        """Helper method to remove all signatures from the image configuration."""
        self.set_signature_list([])

    def get_signature_list(self) -> List:
        """
        Retrieves the signature list from the image configuration.

        Example format:
        [
          { "digest":"sha256:0123456789",
            "signature":"<sigvalue1>"
          },
          { "digest":"sha256:9876543210",
            "signature":"<sigvalue2>"
          },
          ...
        ]

        Returns:
            The deserialized / unescaped signature list in JSON form.
        """
        labels = ImageConfig._get_labels(self.config_json)
        return json.loads(
            labels.get(
                ImageConfig.SIGNATURES_LABEL, ImageConfig.DEFAULT_SIGNATURES_VALUE
            )
        )

    def set_signature_list(self, signatures: List):
        """
        Serializes / escapes and assigns signature list to the image configuration. The method modifies the raw image
        configuration.

        Args:
            signatures: The deserialized / unescaped signature list in JSON form.
        """
        labels = ImageConfig._get_labels(self.config_json)
        labels[ImageConfig.SIGNATURES_LABEL] = json.dumps(signatures)
        self._set_config_json(self.config_json)

    def sign(self, signer: Signer, endorse: bool = False) -> str:
        """
        Signs or endorses the SHA256 digest value of image configuration, in canonical JSON form, and appends it to the
        signature list.

        (Co-)signatures remove all existing signatures before calculating the canonical digest; endorsements do not.
        Co-endorsements are not supported; however, nested endorsements are supported.

        Effectively, this allows for interlacing (co-)signatures and endorsements, where all (co-)signatures apply to a
        single, signature-less, image configuration regardless of where they appear in the signature list. And all
        endorsements apply to the image configuration with signatures order 0 through n-1 (where n is the order of
        endorsement being verified); regardless of what signatures or endorsements were added afterwards.

        Args:
            signer: The signer used to create the signature value.
            endorse: Toggles between signing (default), and endorsing.

        Returns:
            The signature value as defined by :func:~docker_sign_verify.Signers.sign.
        """
        signatures = self.get_signature_list()
        if not endorse:
            self.clear_signature_list()
        digest = self.get_config_digest_canonical()
        signature = signer.sign(digest.encode("utf-8"))
        if not signature:
            raise RuntimeError("Failed to create signature!")

        signatures.append({"digest": digest, "signature": signature})
        self.set_signature_list(signatures)

        return signature

    def verify_signatures(self) -> Dict:
        """
        Verifies the PEM encoded signature values in the image configuration.

        Returns:
            dict:
                signature_data: List as defined by :func:~docker_sign_verify.ImageConfig.get_signature_list.
                results: Signer-specific result value.
        """
        signatures = self.get_signature_list()
        if not signatures:
            raise RuntimeError("Image does not contain any signatures!")

        # Assumptions:
        # * The signature list is ordered.
        # * The first entry in the list *must* be a (co-)signature, as there is nothing older to endorse.
        # * Normalization during canonicalization ensures a consistent empty set.

        results = []
        for i, signature in enumerate(signatures):
            _temp = copy.deepcopy(self)
            # (Co-)signature
            if signature["digest"] == signatures[0]["digest"]:
                _temp.clear_signature_list()
            # Endorsement
            else:
                _temp.set_signature_list(signatures[:i])

            digest = _temp.get_config_digest_canonical()
            must_be_equal(
                signature["digest"], digest, "Image config canonical digest mismatch"
            )

            signer = Signer.for_signature(signature["signature"])
            result = signer.verify(digest.encode("utf-8"), signature["signature"])
            results.append(result)

        return {"signatures": signatures, "results": results}



#!/usr/bin/env python

"""Docker verify command line interface."""

import logging

from typing import List

import click
import urllib3

from docker_sign_verify import (
    ArchiveImageSource,
    DeviceMapperRepositoryImageSource,
    RegistryV2ImageSource,
)

from .common import logging_options, set_log_levels, version
from .utils import to_image_name

LOGGER = logging.getLogger(__name__)

# Bug Fix: There isn't anything we can do about mis-configured remote certificates ...
urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)


def verify_options(function):
    """Common verification options."""
    function = click.argument(
        "images", callback=to_image_name, nargs=-1, required=True
    )(function)
    return function


def verify(context):
    """Verifies an image(s)."""
    for image_name in context.obj["images"]:
        if context.obj["check_signatures"]:
            result = context.obj["imagesource"].verify_image_signatures(image_name)
            LOGGER.info(
                "Image %s (%s) is consistent.",
                image_name,
                result["image_config"].get_config_digest(),
            )
            LOGGER.info("Image %s signature(s) verified.", len(result["signatures"]))
        else:
            result = context.obj["imagesource"].verify_image_integrity(image_name)
            LOGGER.info(
                "Image %s (%s) is consistent.",
                image_name,
                result["image_config"].get_config_digest(),
            )
            LOGGER.info("Image signature(s) not verified.")


@click.group()
@click.option(
    "--check-signatures/--no-check-signatures",
    default=True,
    help="Toggles integrity vs integrity and signature checking.",
    show_default=True,
)
@logging_options
@click.pass_context
def cli(context, check_signatures: bool, verbosity: int = 2):
    """Verifies embedded signatures, and the integrity of docker image layers and metadata."""

    set_log_levels(verbosity)

    context.obj = {"check_signatures": check_signatures}


@cli.command()
@click.option(
    "-a",
    "--archive",
    help="Uncompressed docker image archive.",
    required=True,
    type=click.Path(dir_okay=False, resolve_path=True),
)
@verify_options
@click.pass_context
# pylint: disable=redefined-outer-name
def archive(context, images: List, archive: str):
    """Operates on docker-save produced archives."""

    context.obj["images"] = images
    context.obj["imagesource"] = ArchiveImageSource(archive=archive)
    verify(context)


@cli.command()
@verify_options
@click.pass_context
def registry(context, images: List):
    """Operates on docker registries (v2)."""

    context.obj["images"] = images
    context.obj["imagesource"] = RegistryV2ImageSource()
    verify(context)


@cli.command()
@verify_options
@click.pass_context
def repository(context, images: List):
    """Operates on docker repositories."""

    context.obj["images"] = images
    context.obj["imagesource"] = DeviceMapperRepositoryImageSource()
    verify(context)


cli.add_command(version)

if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    cli()



#!/usr/bin/env python

"""Classes that provide a source of docker images."""

import abc
import copy
import datetime
import hashlib
import io
import json
import logging
import os
import random
import subprocess
import tempfile
import time

from pathlib import Path
from typing import Dict
from urllib.parse import urlparse

import gnupg  # Needed for type checking
import requests
from requests.models import Response
import www_authenticate

from .manifests import (
    ArchiveManifest,
    DeviceMapperRepositoryManifest,
    Manifest,
    RegistryV2Manifest,
)
from .imageconfig import ImageConfig
from .imagename import ImageName
from .signers import Signer
from .utils import (
    copy_file,
    file_exists_in_tar,
    formatted_digest,
    gunzip,
    must_be_equal,
    read_file,
    tar,
    tar_add_file,
    tar_delete_file,
    tar_mkdir,
    untar,
    write_file,
    xellipsis,
    FormattedSHA256,
)

LOGGER = logging.getLogger(__name__)


class ImageSource(abc.ABC):
    """
    Abstract source of docker images.
    """

    def __init__(self, *, dry_run: bool = False):
        """
        Args:
            dry_run: If true, destination image sources will not be changed.
        """
        self.dry_run = dry_run

    def _sign_image_config(self, signer: Signer, image_name: ImageName) -> Dict:
        """
        Verifies an image, then signs it without storing it in the image source.

        Args:
            signer: The signer used to create the signature value.
            image_name: The image name.

        Returns:
            dict:
                image_config: The ImageConfig object corresponding to the signed image.
                signature_value: as defined by :func:~docker_sign_verify.ImageConfig.sign.
                verify_image_data: as defined by :func:~docker_sign_verify.ImageSource.verify_image_integrity.
        """
        # Verify image integrity (we use the verified values from this point on)
        data = self.verify_image_integrity(image_name)
        image_config = data["image_config"]  # type: ImageConfig

        # Append our signature to any existing signatures ....
        signature_value = image_config.sign(signer)

        return {
            "image_config": image_config,
            "signature_value": signature_value,
            "verify_image_data": data,
        }

    def _verify_image_config(self, image_name: ImageName) -> Dict:
        """
        Verifies the integration of an image configuration against metadata contained within a manifest.

        Args:
            image_name: The image name for which to retrieve the configuration.

        Returns:
            dict:
                image_config: The image configuration.
                image_layers: The listing of image layer identifiers.
                manifest: The image-source specific manifest.
                manifest_layers: The listing of manifest layer identifiers.
        """

        # Retrieve the image configuration digest and layers identifiers from the manifest ...
        LOGGER.debug("Verifying Integrity: %s ...", image_name)
        manifest = self.get_manifest(image_name)
        config_digest = manifest.get_config_digest(image_name)
        LOGGER.debug("    config digest: %s", xellipsis(config_digest))
        manifest_layers = manifest.get_layers(image_name)
        LOGGER.debug(
            "    manifest layers:\n\t\t\t%s",
            "\n\t\t\t".join([xellipsis(l) for l in manifest_layers]),
        )

        # Retrieve the image layers from the image configuration ...
        image_config = self.get_image_config(image_name)
        config_digest_canonical = image_config.get_config_digest_canonical()
        LOGGER.debug(
            "    config digest (canonical): %s", xellipsis(config_digest_canonical)
        )
        must_be_equal(
            config_digest,
            image_config.get_config_digest(),
            "Image config digest mismatch",
        )
        image_layers = image_config.get_image_layers()
        LOGGER.debug(
            "    image layers:\n\t\t\t%s",
            "\n\t\t\t".join([xellipsis(l) for l in image_layers]),
        )

        # Quick check: Ensure that the layer counts are consistent
        must_be_equal(len(manifest_layers), len(image_layers), "Layer count mismatch")

        return {
            "image_config": image_config,
            "image_layers": image_layers,
            "manifest": manifest,
            "manifest_layers": manifest_layers,
        }

    @abc.abstractmethod
    def get_image_config(self, image_name: ImageName) -> ImageConfig:
        """
        Retrieves an image configuration (config.json).

        Args:
            image_name: The image name.

        Returns:
            The image configuration.
        """

    @abc.abstractmethod
    # TODO: Can we make layer type 'FormattedSHA256' ???
    def get_image_layer_to_disk(self, image_name: ImageName, layer: str, file):
        """
        Retrieves a single image layer stored to disk.

        Args:
            image_name: The image name.
            layer: The layer identifier in the form: <hash type>:<digest value>.
            file: File in which to store the image layer.
        """

    @abc.abstractmethod
    def get_manifest(self, image_name: ImageName = None) -> Manifest:
        """
        Retrieves the manifest for a given image.

        Args:
            image_name: The name image for which to retrieve the manifest.

        Returns:
            The image source-specific manifest.
        """

    @abc.abstractmethod
    def put_manifest(self, manifest: Manifest, image_name: ImageName = None):
        """
        Assigns the manifest for a given image.

        Args:
            manifest: The image source-specific manifest to be assigned.
            image_name: The name of the image for which to assign the manifest.
        """

    @abc.abstractmethod
    def put_image_config(self, image_name: ImageName, image_config: ImageConfig):
        """
        Assigns an image configuration (config.json).

        Args:
            image_name: The image name.
            image_config: The image configuration to be assigned.
        """

    @abc.abstractmethod
    def put_image_layer(self, image_name: ImageName, content):
        """
        Assigns a single image layer.

        Args:
            image_name: The image name.
            content: The layer content.
        """

    @abc.abstractmethod
    def put_image_layer_from_disk(self, image_name: ImageName, file):
        """
        Assigns a single image layer read from disk.

        Args:
            image_name: The image name.
            file: File from which to read the layer content.
        """

    @abc.abstractmethod
    def layer_exists(self, image_name: ImageName, layer: FormattedSHA256) -> bool:
        """
        Checks if a given image layer exists.

        Args:
            image_name: The image name.
            layer: The layer identifier in the form: <hash type>:<digest value>.

        Returns:
            bool: True if the layer exists, False otherwise.
        """

    @abc.abstractmethod
    def sign_image(
        self,
        signer: Signer,
        src_image_name: ImageName,
        dest_image_source,
        dest_image_name: ImageName,
    ):
        """
        Retrieves, verifies and signs the image, storing it in the destination image source.

        Args:
            signer: The signer used to create the signature value.
            src_image_name: The source image name.
            dest_image_source: The destination image source into which to store the signed image.
            dest_image_name: The description image name.

        Returns:
            dict: as defined by :func:~docker_sign_verify.ImageSource._sign_image_config.
        """

    @abc.abstractmethod
    def verify_image_integrity(self, image_name: ImageName) -> Dict:
        """
        Verifies that the image source data format is consistent with respect to the image configuration and image
        layers, and that the image configuration and image layers are internally consistent (the digest values match).

        Args:
            image_name: The image name.

        Returns:
            dict:
                image config: The image configuration.
                manifest: The image source-specific manifest file (archive, registry, repository).
                uncompressed_layer_files: The list of uncompressed layer files on disk.
        """

    def verify_image_signatures(self, image_name: ImageName) -> Dict:
        """
        Verifies that signatures contained within the image source data format are valid (that the image has not been
        modified since they were created)

        Args:
            image_name: The image name.

        Returns:
            dict:
                image config: The image configuration.
                signatures: as defined by :func:~docker_sign_verify.ImageConfig.verify_signatures.
        """

        # Verify image integrity (we use the verified values from this point on)
        image_config = self.verify_image_integrity(image_name)["image_config"]

        # Verify image signatures ...
        LOGGER.debug("Verifying Signature(s): %s ...", image_name)
        LOGGER.debug(
            "    config digest (signed): %s",
            xellipsis(image_config.get_config_digest()),
        )
        data = image_config.verify_signatures()

        # List the image signatures ...
        LOGGER.debug("    signatures:")
        for result in data["results"]:
            # pylint: disable=protected-access
            if isinstance(result, gnupg._parsers.Verify):
                if not result.valid:
                    raise RuntimeError(
                        "Verification failed for signature with key_id '{0}': {1}".format(
                            result.key_id, result.status
                        )
                    )
                LOGGER.debug(
                    "        Signature made %s using key ID %s",
                    time.strftime(
                        "%Y-%m-%d %H:%M:%S", time.gmtime(float(result.sig_timestamp))
                    ),
                    result.key_id,
                )
                LOGGER.debug("            %s", result.username)
            elif result.get("type", None) == "pki":
                if not result["valid"]:
                    raise RuntimeError(
                        "Verification failed for signature using cert: {0}".format(
                            result["keypair_path"]
                        )
                    )
                # TODO: Add better debug logging
                LOGGER.debug("        Signature made using undetailed PKI keypair.")
            else:
                LOGGER.error("Unknown Signature Type: %s", type(result))

        LOGGER.debug("Signature check passed.")

        return {"image_config": image_config, "signatures": data["results"]}


class ArchiveImageSource(ImageSource):
    """
    Docker archive image source.
    """

    FILE_ARCHIVE_MANIFEST = "manifest.json"
    CHUNK_SIZE = 4096

    def __init__(self, *, archive, **kwargs):
        """
        Args:
            archive: Path to the docker image archive.
        """
        super(ArchiveImageSource, self).__init__(**kwargs)
        self.archive = archive

    def _file_exists(self, name) -> bool:
        """
        Checks if a give file exists within this image source.

        Args:
            name: Name of the file to be checked.

        Returns:
            bool: True if the file exists, False otherwise.
        """
        with open(self.archive, "rb") as file_in:
            return file_exists_in_tar(file_in, name)

    def get_file_from_archive(self, path):
        """
        Retrieves a file from within this image source.

        Args:
            path: Relative path of the file to be retrieved.

        Returns:
            The file content.
        """
        bytesio = io.BytesIO()
        with open(self.archive, "rb") as file:
            untar(file, path, bytesio)
        return bytesio.read()

    # ImageSource Members

    def get_image_config(self, image_name: ImageName) -> ImageConfig:
        config = self.get_manifest().get_config(image_name)
        return ImageConfig(self.get_file_from_archive(config["Config"]))

    def get_image_layer_to_disk(self, image_name: ImageName, layer: str, file):
        with open(self.archive, "rb") as file_in:
            return untar(file_in, ArchiveManifest.digest_to_layer(layer), file)

            #    def put_image_layer(self, image, content):

    def get_manifest(self, image_name: ImageName = None) -> Manifest:
        raw_archive_manifest = self.get_file_from_archive(
            ArchiveImageSource.FILE_ARCHIVE_MANIFEST
        )
        return ArchiveManifest(raw_archive_manifest)

    def put_image_config(self, image_name: ImageName, image_config: ImageConfig):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_image_config")
            return
        digest = image_config.get_config_digest()
        name = "{0}.json".format(digest.sha256)
        if not self._file_exists(name):
            with open(self.archive, "rb+") as file_out:
                tar_add_file(file_out, name, image_config.get_config())

    def put_image_layer(self, image_name: ImageName, content):
        raise NotImplementedError

    def put_image_layer_from_disk(self, image_name: ImageName, file) -> FormattedSHA256:
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_image_layer_from_disk")
            return FormattedSHA256("0" * 64)
        # TODO: Do we really want to use random garbage here???
        #       Look into moby/.../save.go to find what to use instead.
        digest = formatted_digest(
            "{0}{1}{2}".format(
                str(image_name), datetime.datetime.now(), random.randint(1, 101)
            ).encode("utf-8")
        )
        layer = ArchiveManifest.digest_to_layer(digest)
        with open(self.archive, "rb+") as file_out:
            tar_mkdir(file_out, os.path.dirname(layer))
            file_out.seek(0)
            tar(file_out, layer, file)
        return digest

    def put_manifest(self, manifest: Manifest, image_name: ImageName = None):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_manifest")
            return
        with open(self.archive, "rb+") as file_out:
            tar_delete_file(file_out, ArchiveImageSource.FILE_ARCHIVE_MANIFEST)
            file_out.seek(0)
            tar_add_file(
                file_out,
                ArchiveImageSource.FILE_ARCHIVE_MANIFEST,
                str(manifest).encode("utf-8"),
            )

    def layer_exists(self, image_name: ImageName, layer: FormattedSHA256) -> bool:
        return self._file_exists(ArchiveManifest.digest_to_layer(layer))

    def sign_image(
        self,
        signer: Signer,
        src_image_name: ImageName,
        dest_image_source: ImageSource,
        dest_image_name: ImageName,
    ):
        LOGGER.debug("Signing: %s ...", src_image_name)

        # Generate a signed image configuration ...
        data = self._sign_image_config(signer, src_image_name)
        manifest = data["verify_image_data"]["manifest"]
        LOGGER.debug("    Signature:\n%s", data["signature_value"])
        image_config = data["image_config"]

        # Replicate all of the image layers ...
        archive_layers = manifest.get_layers(src_image_name)
        archive_layers_changed = archive_layers.copy()
        for i, archive_layer in enumerate(archive_layers):
            if not dest_image_source.layer_exists(dest_image_name, archive_layer):
                # Update the layer
                digest = dest_image_source.put_image_layer_from_disk(
                    dest_image_name,
                    data["verify_image_data"]["uncompressed_layer_files"][i],
                )
                archive_layers_changed[i] = digest
        archive_layers = archive_layers_changed

        # Push the new image configuration ...
        config_digest_signed = image_config.get_config_digest()
        LOGGER.debug("    config digest (signed): %s", config_digest_signed)
        dest_image_source.put_image_config(dest_image_name, image_config)

        # Generate a new archive manifest, and push ...
        if isinstance(dest_image_source, ArchiveImageSource):
            manifest_signed = dest_image_source.get_manifest()  # type: ArchiveManifest

            repotags = None
            if dest_image_name.tag:
                repotags = [str(dest_image_name)]
                manifest_signed.append_config(
                    config_digest_signed, archive_layers, repotags
                )
            data["manifest_signed"] = manifest_signed
            # TODO: make sure to remove conflicting tags in "other" config entries
            dest_image_source.put_manifest(manifest_signed)
            # TODO: Update foo.tar:/repositories as well
        elif isinstance(dest_image_source, RegistryV2ImageSource):
            raise NotImplementedError
        elif isinstance(dest_image_source, DeviceMapperRepositoryImageSource):
            raise NotImplementedError
        else:
            raise RuntimeError(
                "Unknown derived class: {0}".format(type(dest_image_source))
            )

        LOGGER.debug("Created new image: %s", dest_image_name)

        return data

    def verify_image_integrity(self, image_name: ImageName) -> Dict:
        data = self._verify_image_config(image_name)

        # Reconcile manifest layers and image layers (in order)...
        uncompressed_layer_files = []
        for i, layer in enumerate(data["manifest_layers"]):
            # Retrieve the archive image layer and verify the digest ...
            uncompressed_layer_files.append(tempfile.NamedTemporaryFile())
            data_uncompressed = self.get_image_layer_to_disk(
                image_name, layer, uncompressed_layer_files[i]
            )
            must_be_equal(
                data["image_layers"][i],
                data_uncompressed["digest"],
                "Archive layer[{0}] digest mismatch".format(i),
            )

        LOGGER.debug("Integrity check passed.")

        return {
            "image_config": data["image_config"],
            "manifest": data["manifest"],
            "uncompressed_layer_files": uncompressed_layer_files,
        }


class RegistryV2ImageSource(ImageSource):
    """
    Docker registry image source.
    """

    BLOB_MIME_TYPE = "application/vnd.docker.container.image.v1+json"
    BLOB_URL_PATTERN = "https://{0}/v2/{1}/blobs/{2}"
    DOCKERHUB_AUTH_URL_PATTERN = (
        "{0}?service={1}&scope={2}&client_id=docker-sign-verify"
    )
    MANIFEST_MIME_TYPE = "application/vnd.docker.distribution.manifest.v2+json"
    MANIFEST_URL_PATTERN = "https://{0}/v2/{1}/manifests/{2}"

    DEFAULT_CREDENTIALS_STORE = Path.home().joinpath(".docker/config.json")

    CHUNK_SIZE = 4096

    def __init__(self, *, credentials_store=None, **kwargs):
        """
        Args:
            credentials_store: Path to the docker registry credentials store.
        """
        super(RegistryV2ImageSource, self).__init__(**kwargs)
        if not credentials_store:
            credentials_store = os.environ.get(
                "DSV_CREDENTIALS_STORE", RegistryV2ImageSource.DEFAULT_CREDENTIALS_STORE
            )
        self.credentials_store = credentials_store
        self.credentials = None
        self.token = None

    def _get_auth_token(self, credentials: str, endpoint: str, scope: str) -> str:
        """
        Retrieves the registry auth token for a given scope.

        Args:
            credentials: The credentials to use to retrieve the auth token.
            endpoint: Registry endpoint for which to retrieve the token.
            scope: The scope of the auth token.

        Returns:
            The corresponding auth token, or None.
        """

        # https://github.com/docker/distribution/blob/master/docs/spec/auth/token.md
        if not self.token:
            # Test using HTTP basic authentication to retrieve the www-authenticate response header ...
            headers = {"Authorization": "Basic {0}".format(credentials)}
            url = "https://{0}/v2/".format(endpoint)
            response = requests.get(url, headers=headers)

            auth_params = www_authenticate.parse(response.headers["Www-Authenticate"])
            bearer = auth_params["bearer"]

            url = RegistryV2ImageSource.DOCKERHUB_AUTH_URL_PATTERN.format(
                bearer["realm"], bearer["service"], scope
            )
            response = requests.get(url, headers=headers)
            # LOGGER.debug("Token Response: %s", response.content)
            must_be_equal(200, response.status_code, "Failed to retrieve bearer token")

            self.token = response.json()["token"]

        return self.token

    def _get_credentials(self, endpoint: str) -> str:
        """
        Retrieves the registry credentials from the docker registry credentials store for a given endpoint

        Args:
            endpoint: Registry endpoint for which to retrieve the credentials.

        Returns:
            The corresponding base64 encoded registry credentials, or None.
        """
        result = None

        if not self.credentials and self.credentials_store:
            LOGGER.debug("Using credentials store: %s", self.credentials_store)

            # TODO: Add support for secure providers:
            #       https://docs.docker.com/engine/reference/commandline/login/#credentials-store

            self.credentials = {}
            if self.credentials_store.is_file():
                with self.credentials_store.open(mode="rb") as file:
                    self.credentials = json.loads(file.read()).get("auths", {})

        for endpoint_auth in [
            u for u in self.credentials if endpoint in (u, urlparse(u).netloc)
        ]:
            result = self.credentials[endpoint_auth].get("auth", None)
            if result:
                break

        return result

    def _get_request_headers(self, image_name: ImageName, headers=None):
        """
        Generates request headers that contain registry credentials for a given registry endpoint.

        Args:
            image_name: Image name for which to retrieve the request headers.
            headers: Optional supplemental request headers to be returned.

        Returns:
            The generated request headers.
        """
        if not headers:
            headers = {}

        if "User-Agent" not in headers:
            # Note: This cannot be imported above, as it causes a circular import!
            from . import __version__

            headers["User-Agent"] = "docker-sign-verify/{0}".format(__version__)

        endpoint = image_name.resolve_endpoint()
        credentials = self._get_credentials(endpoint)
        if credentials:
            if "docker.io" in endpoint:
                token = self._get_auth_token(
                    credentials,
                    endpoint,
                    "repository:{0}:pull".format(image_name.resolve_image()),
                )
                headers["Authorization"] = "Bearer {0}".format(token)
            else:
                headers["Authorization"] = "Basic {0}".format(credentials)

        return headers

    def monolithic_blob_upload(
        self, image_name: ImageName, blob, digest: FormattedSHA256 = None
    ) -> Response:
        """
        Performs a monolithic upload of an image blob.

        Args:
            image_name: The name of the target repository.
            blob: The data to be uploaded.
            digest: Digest of uploaded blob.

        Returns:
            The underlying response.
        """
        url = RegistryV2ImageSource.BLOB_URL_PATTERN.format(
            image_name.resolve_endpoint(), image_name.resolve_image(), "uploads/"
        )
        headers = self._get_request_headers(
            image_name, {"Content-Type": "application/octet-stream"}
        )
        if not digest:
            digest = formatted_digest(blob)
        response = requests.post(
            url, params={"digest": digest}, headers=headers, data=blob
        )
        must_be_equal(202, response.status_code, "Failed to upload blob")

        # TODO: Do we need to emulate the Docker-Content-Digest return type?
        return response

    def initiate_resumable_blob_upload(self, image_name: ImageName) -> Response:
        """
        Starts a resumable upload of an image blob.

        Args:
            image_name: The name of the target repository.

        Returns:
            The underlying response.
        """
        url = RegistryV2ImageSource.BLOB_URL_PATTERN.format(
            image_name.resolve_endpoint(), image_name.resolve_image(), "uploads/"
        )
        headers = self._get_request_headers(image_name)
        response = requests.post(url, headers=headers)
        # logging.debug("UUID: {0}".format(response.headers["Docker-Upload-UUID"]))
        must_be_equal(
            202, response.status_code, "Failed to initiate resumable blob upload"
        )
        return response

    def resume_blob_upload(self, location, offset, chunk) -> Response:
        """
        Continues a resumable upload of an image blob.

        Args:
            location:
                The Resumable location provided from
                :func:~docker_sign_verify.RegistryV2ImageSource.initiate_resumable_blob_upload.
            offset: Range of bytes identifying the desired block of content represented by the body.
            chunk: The chunk of data to be uploaded.

        Returns:
            The underlying response.
        """
        headers = self._get_request_headers(
            ImageName.parse(urlparse(location).netloc + "/"),
            {
                "Content-Range": "{0}-{1}".format(offset, offset + len(chunk) - 1),
                "Content-Type": "application/octet-stream",
            },
        )
        response = requests.patch(location, headers=headers, data=chunk)
        must_be_equal(202, response.status_code, "Failed to upload blob chunk")

        return response

    def complete_resumable_blob_upload(
        self, location, digest: FormattedSHA256, chunk=None
    ) -> FormattedSHA256:
        """
        Finishes a resumable upload of an image blob.

        Args:
            location:
                The Resumable location provided from
                :func:~docker_sign_verify.RegistryV2ImageSource.initiate_resumable_blob_upload.
            digest: Digest of uploaded blob.
            chunk: The chunk of data to be uploaded.

        Returns:
            FormattedSHA256: Digest of the targeted content for the request.
        """
        headers = self._get_request_headers(
            ImageName.parse(urlparse(location).netloc + "/"),
            {"Content-Type": "application/octet-stream"},
        )
        response = requests.put(
            location, params={"digest": digest}, headers=headers, data=chunk
        )
        must_be_equal(
            201, response.status_code, "Failed to complete resumable blob upload"
        )

        return FormattedSHA256.parse(response.headers["Docker-Content-Digest"])

    # TODO: Re-implement or remove this and use get_image_layer_to_disk with BytesIO instead.
    def get_image_layer(self, image_name: ImageName, layer: FormattedSHA256):
        """
        Retrieves a single image layer.

        Args:
            image_name: Name of the image under which the layer exists.
            layer: The layer identifier in the form: <hash type>:<digest value>.

        Returns:
            TODO
        """
        headers = self._get_request_headers(
            image_name, {"Accept": RegistryV2ImageSource.BLOB_MIME_TYPE}
        )
        url = RegistryV2ImageSource.BLOB_URL_PATTERN.format(
            image_name.resolve_endpoint(), image_name.resolve_image(), layer
        )
        response = requests.get(url, headers=headers)
        response.raise_for_status()
        return response.content

    # ImageSource Members

    def get_image_config(self, image_name: ImageName) -> ImageConfig:
        config_digest = self.get_manifest(image_name).get_config_digest()
        return ImageConfig(self.get_image_layer(image_name, config_digest))

    def get_image_layer_to_disk(self, image_name: ImageName, layer: str, file):
        headers = self._get_request_headers(
            image_name, {"Accept": RegistryV2ImageSource.BLOB_MIME_TYPE}
        )
        url = RegistryV2ImageSource.BLOB_URL_PATTERN.format(
            image_name.resolve_endpoint(), image_name.resolve_image(), layer
        )
        response = requests.get(url, headers=headers, stream=True)

        hasher = hashlib.sha256()
        for chunk in response.iter_content(chunk_size=RegistryV2ImageSource.CHUNK_SIZE):
            if not chunk:
                break
            file.write(chunk)
            hasher.update(chunk)
        file.flush()
        os.fsync(file.fileno())

        size = file.tell()

        # Be kind, rewind ...
        file.seek(0)

        return {"digest": FormattedSHA256(hasher.hexdigest()), "size": size}

    def get_manifest(self, image_name: ImageName = None) -> Manifest:
        headers = self._get_request_headers(
            image_name, {"Accept": RegistryV2ImageSource.MANIFEST_MIME_TYPE}
        )
        url = RegistryV2ImageSource.MANIFEST_URL_PATTERN.format(
            image_name.resolve_endpoint(), image_name.resolve_image(), image_name.tag
        )

        response = requests.get(url, headers=headers)
        response.raise_for_status()
        raw_registry_manifest = response.content

        return RegistryV2Manifest(raw_registry_manifest)

    def put_image_config(self, image_name: ImageName, image_config: ImageConfig):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_image_config")
            return
        if not self.layer_exists(image_name, image_config.get_config_digest()):
            self.put_image_layer(image_name, image_config.get_config())

    def put_image_layer(self, image_name: ImageName, content):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_image_layer")
            return

        # TODO: Why doesn't this work?
        # self.monolithic_blob_upload(image_name, content)

        digest = formatted_digest(content)
        location = self.initiate_resumable_blob_upload(image_name).headers["location"]
        return self.complete_resumable_blob_upload(location, digest, content)

        # TODO: Should we check self.layer_exists(image_name, digest) here, like docker CLI does?

    def put_image_layer_from_disk(self, image_name: ImageName, file):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_image_layer_from_disk")
            return
        location = self.initiate_resumable_blob_upload(image_name).headers["location"]

        file.seek(0)
        hasher = hashlib.sha256()
        while True:
            offset = file.tell()
            chunk = file.read(RegistryV2ImageSource.CHUNK_SIZE)
            if not chunk:
                break
            location = self.resume_blob_upload(location, offset, chunk).headers[
                "location"
            ]
            hasher.update(chunk)
        digest = FormattedSHA256(hasher.hexdigest())

        return self.complete_resumable_blob_upload(location, digest)

        # TODO: Should we check self.layer_exists(image_name, digest) here, like docker CLI does?

    def put_manifest(self, manifest: Manifest, image_name: ImageName = None):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_manifest")
            return
        url = RegistryV2ImageSource.MANIFEST_URL_PATTERN.format(
            image_name.resolve_endpoint(), image_name.resolve_image(), image_name.tag
        )
        headers = self._get_request_headers(
            image_name, {"Content-Type": RegistryV2ImageSource.MANIFEST_MIME_TYPE}
        )

        response = requests.put(
            url, headers=headers, data=str(manifest).encode("utf-8")
        )
        must_be_equal(201, response.status_code, "Failed to upload manifest")

        return response

    def layer_exists(self, image_name: ImageName, layer: FormattedSHA256) -> bool:
        url = RegistryV2ImageSource.BLOB_URL_PATTERN.format(
            image_name.resolve_endpoint(), image_name.resolve_image(), layer
        )
        headers = self._get_request_headers(image_name)
        response = requests.head(url, headers=headers)

        return response.status_code in [200, 307]

    def sign_image(
        self,
        signer: Signer,
        src_image_name: ImageName,
        dest_image_source: ImageSource,
        dest_image_name: ImageName,
    ):
        LOGGER.debug("Signing: %s ...", src_image_name)

        # Generate a signed image configuration ...
        data = self._sign_image_config(signer, src_image_name)
        manifest = data["verify_image_data"]["manifest"]
        LOGGER.debug("    Signature:\n%s", data["signature_value"])
        image_config = data["image_config"]

        # Replicate all of the image layers ...
        registry_layers = manifest.get_layers()
        for i, registry_layer in enumerate(registry_layers):
            if not dest_image_source.layer_exists(dest_image_name, registry_layer):
                dest_image_source.put_image_layer_from_disk(
                    dest_image_name,
                    data["verify_image_data"]["compressed_layer_files"][i],
                )

        # Push the new image configuration ...
        config_digest_signed = image_config.get_config_digest()
        LOGGER.debug("    config digest (signed): %s", config_digest_signed)
        dest_image_source.put_image_config(dest_image_name, image_config)

        # Generate a new registry manifest, and push ...
        if isinstance(dest_image_source, ArchiveImageSource):
            raise NotImplementedError
        elif isinstance(dest_image_source, RegistryV2ImageSource):
            manifest_signed = copy.deepcopy(manifest)  # type: RegistryV2Manifest
            manifest_signed.set_config_digest(
                config_digest_signed, len(image_config.get_config())
            )
            data["manifest_signed"] = manifest_signed
            dest_image_source.put_manifest(manifest_signed, dest_image_name)
        elif isinstance(dest_image_source, DeviceMapperRepositoryImageSource):
            raise NotImplementedError
        else:
            raise RuntimeError(
                "Unknown derived class: {0}".format(type(dest_image_source))
            )

        LOGGER.debug("Created new image: %s", dest_image_name)

        return data

    def verify_image_integrity(self, image_name: ImageName) -> Dict:
        data = self._verify_image_config(image_name)

        # Reconcile manifest layers and image layers (in order)...
        compressed_layer_files = []
        uncompressed_layer_files = []
        for i, layer in enumerate(data["manifest_layers"]):
            # Retrieve the registry image layer and verify the digest ...
            compressed_layer_files.append(tempfile.NamedTemporaryFile())
            data_compressed = self.get_image_layer_to_disk(
                image_name, layer, compressed_layer_files[i]
            )
            must_be_equal(
                layer,
                data_compressed["digest"],
                "Registry layer[{0}] digest mismatch".format(i),
            )
            must_be_equal(
                os.path.getsize(compressed_layer_files[i].name),
                data_compressed["size"],
                "Registry layer[{0}] size mismatch".format(i),
            )

            # Decompress (convert) the registry image layer into the image layer
            # and verify the digest ...
            uncompressed_layer_files.append(tempfile.NamedTemporaryFile())
            data_uncompressed = gunzip(
                compressed_layer_files[i], uncompressed_layer_files[i]
            )
            must_be_equal(
                data["image_layers"][i],
                data_uncompressed["digest"],
                "Image layer[{0}] digest mismatch".format(i),
            )

        LOGGER.debug("Integrity check passed.")

        return {
            "compressed_layer_files": compressed_layer_files,
            "image_config": data["image_config"],
            "manifest": data["manifest"],
            "uncompressed_layer_files": uncompressed_layer_files,
        }


class DeviceMapperRepositoryImageSource(ImageSource):
    """
    Docker repository image source.
    """

    DOCKER_ROOT = Path("/var/lib/docker")
    DM_CONTENT_ROOT = DOCKER_ROOT.joinpath("image/devicemapper/imagedb/content/sha256")
    DM_LAYER_ROOT = DOCKER_ROOT.joinpath("image/devicemapper/layerdb/sha256")
    DM_METADATA_ROOT = DOCKER_ROOT.joinpath("devicemapper/metadata")
    DM_REPOSITORIES = DOCKER_ROOT.joinpath("image/devicemapper/repositories.json")

    CHUNK_SIZE = 4096

    def __init__(self, **kwargs):
        super(DeviceMapperRepositoryImageSource, self).__init__(**kwargs)

    # ImageSource Members

    def get_image_config(self, image_name: ImageName) -> ImageConfig:
        config_digest = self.get_manifest().get_config_digest(image_name)
        path = DeviceMapperRepositoryImageSource.DM_CONTENT_ROOT.joinpath(
            config_digest.sha256
        )
        raw_image_config = read_file(path)
        return ImageConfig(raw_image_config)

    def get_image_layer_to_disk(self, image_name: ImageName, layer: str, file):
        # Retrieve the devicemapper metadata ...
        path = DeviceMapperRepositoryImageSource.DM_LAYER_ROOT.joinpath(
            layer[7:], "cache-id"
        )
        cache_id = read_file(path).decode("utf-8")
        if not cache_id:
            raise RuntimeError("Unable to find cache id for layer: {0}".format(layer))
        path = DeviceMapperRepositoryImageSource.DM_METADATA_ROOT.joinpath(cache_id)
        raw_metadata = read_file(path)
        metadata = json.loads(raw_metadata)
        if not (metadata["device_id"] or metadata["size"]):
            raise RuntimeError(
                "Unable to find device id and / or size for layer: {0}".format(layer)
            )

        # Create the devicemapper table ...
        device_name = "dsv-{0}".format(layer)
        # TODO: How to we find the vgname?
        volume_group = "/dev/mapper/rhel-docker--pool"
        table = "0 {0} thin {1} {2}".format(
            metadata["size"], volume_group, metadata["device_id"]
        )
        subprocess.run(
            ["/sbin/dmsetup", "create", device_name, "--table", table], check=True
        )

        # Mount the layer ...
        mount = Path(tempfile.mkdtemp())
        subprocess.run(
            [
                "mount",
                "-o",
                "ro",
                Path("/dev/mapper").joinpath(device_name),
                mount.absolute(),
            ],
            check=True,
        )

        # Reconstruct the layer tar ...
        path = DeviceMapperRepositoryImageSource.DM_LAYER_ROOT.joinpath(
            layer[7:], "tar-split.json.gz"
        )
        rootfs = mount.joinpath("rootfs")
        if not rootfs.exists():
            raise RuntimeError(
                "Root filesystem does not exist for layer: {0}".format(layer)
            )
        # TODO: Replace this with a pure-python implementation
        process = subprocess.Popen(
            ["tar-split", "asm", "--input", str(path), "--path", rootfs],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )
        # TODO: This still buffers the entire file in memory ...
        result = copy_file(process.stdout, file)
        return_code = process.wait(120)
        if return_code != 0:
            raise RuntimeError("Tar-split failed!")

        # Unmount the layer, remove the devicemapper table, delete the mountpoint ...
        subprocess.run(["umount", mount.absolute()], check=True)
        subprocess.run(["/sbin/dmsetup", "remove", device_name], check=True)
        mount.rmdir()

        return result

    def get_manifest(self, image_name: ImageName = None) -> Manifest:
        raw_repository_manifest = read_file(
            DeviceMapperRepositoryImageSource.DM_REPOSITORIES
        )
        return DeviceMapperRepositoryManifest(raw_repository_manifest)

    def put_image_config(self, image_name: ImageName, image_config: ImageConfig):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_image_config")
            return
        # TODO: Remove debug code
        # path = DeviceMapperRepositoryImageSource.DM_CONTENT_ROOT.joinpath(image_config.get_config_digest().sha256)
        path = Path(
            "/tmp/docker-ridavis/image/devicemapper/imagedb/conent/sha256/"
        ).joinpath(image_config.get_config_digest().sha256)
        path.parent.mkdir(exist_ok=True, parents=True)
        write_file(path, image_config.get_config())

    def put_image_layer(self, image_name: ImageName, content):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_image_layer")
            return
        # TODO: Implement this method ...
        raise NotImplementedError

    def put_image_layer_from_disk(self, image_name: ImageName, file):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_image_layer_from_disk")
            return
        # TODO: Implement this method ...
        raise NotImplementedError

    def put_manifest(self, manifest: Manifest, image_name: ImageName = None):
        if self.dry_run:
            LOGGER.debug("Dry Run: skipping put_manifest")
            return
        # TODO: Remove debug code
        # path = DeviceMapperRepositoryImageSource.DM_REPOSITORIES
        path = Path("/tmp/docker-ridavis/repositories.json")
        path.parent.mkdir(exist_ok=True, parents=True)
        write_file(path, str(manifest).encode("utf-8"))

    def layer_exists(self, image_name: ImageName, layer: FormattedSHA256) -> bool:
        result = False
        path_layer = DeviceMapperRepositoryImageSource.DM_LAYER_ROOT.joinpath(
            layer.sha256
        ).joinpath("cache-id")
        if path_layer.exists():
            cache_id = read_file(path_layer).decode("utf-8")
            path_cache = DeviceMapperRepositoryImageSource.DM_METADATA_ROOT.joinpath(
                cache_id
            )
            result = path_cache.exists()
        return result

    def sign_image(
        self,
        signer: Signer,
        src_image_name: ImageName,
        dest_image_source: ImageSource,
        dest_image_name: ImageName,
    ):
        LOGGER.debug("Signing: %s ...", src_image_name)

        # Generate a signed image configuration ...
        data = self._sign_image_config(signer, src_image_name)
        manifest = data["verify_image_data"]["manifest"]
        LOGGER.debug("    Signature:\n%s", data["signature_value"])
        image_config = data["image_config"]

        # Replicate all of the image layers ...
        repository_layers = manifest.get_layers(src_image_name)
        for i, repository_layer in enumerate(repository_layers):
            if not dest_image_source.layer_exists(dest_image_name, repository_layer):
                dest_image_source.put_image_layer_from_disk(
                    dest_image_name,
                    data["verify_image_data"]["compressed_layer_files"][i],
                )
        # TODO: We we need to track the layer translations here? Is this possible for DM repos?

        # Push the new image configuration ...
        config_digest_signed = image_config.get_config_digest()
        LOGGER.debug("    config digest (signed): %s", config_digest_signed)
        dest_image_source.put_image_config(dest_image_name, image_config)

        # Generate a new repository manifest, and push ...
        if isinstance(dest_image_source, ArchiveImageSource):
            raise NotImplementedError
        elif isinstance(dest_image_source, RegistryV2ImageSource):
            raise NotImplementedError
        elif isinstance(dest_image_source, DeviceMapperRepositoryImageSource):
            manifest_signed = (
                dest_image_source.get_manifest()
            )  # type: DeviceMapperRepositoryManifest

            if dest_image_name.tag:
                manifest_signed.override_config(config_digest_signed, dest_image_name)
            data["manifest_signed"] = manifest_signed
            dest_image_source.put_manifest(manifest_signed, dest_image_name)
        else:
            raise RuntimeError(
                "Unknown derived class: {0}".format(type(dest_image_source))
            )

        LOGGER.debug("Created new image: %s", dest_image_name)

        return data

    def verify_image_integrity(self, image_name: ImageName) -> Dict:
        data = self._verify_image_config(image_name)

        # Note: We do not need to reconcile manifest layer ids here, as "we" derived them in
        # :func:docker_sign_verify.manifests.DeviceMapperRepositoryManifest.get_layers.

        # Reconcile manifest layers and image layers (in order)...
        uncompressed_layer_files = []
        for i, layer in enumerate(data["manifest_layers"]):
            # Retrieve the repository image layer and verify the digest ...
            uncompressed_layer_files.append(tempfile.NamedTemporaryFile())
            data_compressed = self.get_image_layer_to_disk(
                image_name, layer, uncompressed_layer_files[i]
            )
            must_be_equal(
                data["image_layers"][i],
                data_compressed["digest"],
                "Repository layer[{0}] digest mismatch".format(i),
            )

        LOGGER.debug("Integrity check passed.")

        return {
            "image_config": data["image_config"],
            "manifest": data["manifest"],
            "uncompressed_layer_files": uncompressed_layer_files,
        }



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""Signer tests."""

import os
import tempfile
import shutil

from pathlib import Path

import pytest

from docker_sign_verify import GPGSigner, PKISigner


@pytest.fixture()
def gpgsigner(request):
    """Provides a GPGSinger instance."""
    homedir = Path(tempfile.mkdtemp())

    def _remove_homedir():
        shutil.rmtree(homedir)

    request.addfinalizer(_remove_homedir)
    signer = GPGSigner(None, "testing", homedir)
    # pylint: disable=protected-access
    signer._debug_init_store()

    return signer


@pytest.fixture()
def pkisigner(request):
    """Provides a PKISigner instance."""
    keypair = tempfile.mkstemp()[1]

    def _remove_keypair():
        os.remove(keypair)

    request.addfinalizer(_remove_keypair)
    signer = PKISigner(keypair, "testing")
    # pylint: disable=protected-access
    signer._debug_init_keypair()
    return signer


# def test_gpgsigner(caplog, gpgsigner: GPGSigner):
def test_gpgsigner(gpgsigner: GPGSigner):
    """Test configuration signing and verification using GPG."""

    # TODO: Figure out why this isn't working ...
    # caplog.set_level(logging.FATAL, logger="gnupg")

    data = b"TEST DATA"

    # Generate a signature for the test data ...
    signature = gpgsigner.sign(data)
    assert "PGP SIGNATURE" in signature

    # Verify the generated signature against the test data ...
    result = gpgsigner.verify(data, signature)
    assert result.valid
    assert gpgsigner.keyid == result.key_id
    assert result.status == "signature valid"


def test_pkisigner(pkisigner: PKISigner):
    """Test configuration signing and verification using PKI."""

    data = b"TEST DATA"

    # Generate a signature for the test data ...
    signature = pkisigner.sign(data)
    assert "PKI SIGNATURE" in signature

    # Verify the generated signature against the test data ...
    result = pkisigner.verify(data, signature)
    assert result["valid"]
    assert result["keypair_path"] == pkisigner.keypair_path
    assert result["type"] == "pki"



#!/usr/bin/env python

import os
import re

from setuptools import setup, find_packages


def find_version(*segments):
    root = os.path.abspath(os.path.dirname(__file__))
    abspath = os.path.join(root, *segments)
    with open(abspath, "r") as file:
        content = file.read()
    match = re.search(r"^__version__ = ['\"]([^'\"]+)['\"]", content, re.MULTILINE)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string!")


setup(
    author="Richard Davis",
    author_email="crashvb@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    description="A utility that can be used to sign and verify docker images.",
    entry_points="""
        [console_scripts]
        docker-sign=docker_sign_verify.scripts.docker_sign:cli
        docker-verify=docker_sign_verify.scripts.docker_verify:cli
    """,
    extras_require={"dev": ["black", "pylint", "twine", "pytest", "wheel"]},
    include_package_data=True,
    install_requires=[
        "canonicaljson",
        "click",
        "gnupg",
        "pycryptodome",
        "requests",
        "www_authenticate",
    ],
    keywords="docker docker-sign docker-verify integrity sign signatures verify",
    license="Apache License 2.0",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    name="docker_sign_verify",
    packages=find_packages(),
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    test_suite="tests",
    url="https://pypi.org/project/docker-sign-verify/",
    version=find_version("docker_sign_verify", "__init__.py"),
)



#!/usr/bin/env python

# pylint: disable=redefined-outer-name

"""ImageConfig tests."""

import copy

from typing import List

import pytest

from docker_sign_verify import FormattedSHA256, ImageConfig, Signer

from .stubs import _signer_for_signature, FakeSigner
from .testutils import get_test_data


@pytest.fixture
def json_bytes(request) -> bytes:
    """Provides a sample image configuration."""
    return get_test_data(request, __name__, "config.json")


@pytest.fixture
def json_bytes_canonical(request) -> bytes:
    """Provides the canonical form of the sample image configuration."""
    return get_test_data(request, __name__, "config_canonical.json")


@pytest.fixture
def json_bytes_signed(request) -> bytes:
    """Provides a sample image configuration with a single PGP signature."""
    return get_test_data(request, __name__, "config_signed.json")


@pytest.fixture
def json_bytes_signed_canonical(request) -> bytes:
    """Provides the canonical form of the signed sample image configuration."""
    return get_test_data(request, __name__, "config_signed_canonical.json")


@pytest.fixture
def config_digest(request) -> FormattedSHA256:
    """Provides the digest value of the sample image configuration."""
    return FormattedSHA256.parse(
        get_test_data(request, __name__, "config.json.digest", "r")
    )


@pytest.fixture
def config_digest_canonical(request) -> FormattedSHA256:
    """Provides the digest value of canonical form of the sample image configuration."""
    return FormattedSHA256.parse(
        get_test_data(request, __name__, "config_canonical.json.digest", "r")
    )


@pytest.fixture
def config_digest_signed(request) -> FormattedSHA256:
    """"Provides the digest value of the signed sample image configuration."""
    return FormattedSHA256.parse(
        get_test_data(request, __name__, "config_signed.json.digest", "r")
    )


@pytest.fixture
def config_digest_signed_canonical(request) -> FormattedSHA256:
    """Provides the digest value of the canonical form of the signed sample image configuration."""
    return FormattedSHA256.parse(
        get_test_data(request, __name__, "config_signed_canonical.json.digest", "r")
    )


@pytest.fixture
def image_layers() -> List:
    """Provides the list of layer identifiers contained in the sample image configuration."""
    return [
        "sha256:94b2db70f7476c98f4c4a1b7a922136e0c5600d2d74905407ad364dcca2bf852",
        "sha256:22426f366c51f26105aa9a6c6c9aea9fff0f21b7aabfc97870727577edaa3260",
    ]


@pytest.fixture
def image_config(json_bytes: bytes) -> ImageConfig:
    """Provides an ImageConfig instance for the sample image configuration."""
    # Do not use caching; get a new instance for each test
    return ImageConfig(json_bytes)


@pytest.fixture
def image_config_signed(json_bytes_signed: bytes) -> ImageConfig:
    """Provides an ImageConfig instance for the signed sample image configuration."""
    # Do not use caching; get a new instance for each test
    return ImageConfig(json_bytes_signed)


@pytest.fixture
def signature(request) -> str:
    """Provides the PGP signature value used in the signed sample image configuration."""
    return get_test_data(request, __name__, "signature", "r")


def test_init(image_config: ImageConfig, image_config_signed: ImageConfig):
    """Test that signed and unsigned configurations can be instantiated."""
    assert image_config
    assert image_config_signed


def test_str(
    image_config: ImageConfig,
    image_config_signed: ImageConfig,
    json_bytes: bytes,
    json_bytes_signed: bytes,
):
    """Test __str__ pass-through for signed and unsigned configurations (with encoding)."""
    assert str(image_config) == json_bytes.decode("utf-8")
    assert str(image_config_signed) == json_bytes_signed.decode("utf-8")


def test_get_config(
    image_config: ImageConfig,
    image_config_signed: ImageConfig,
    json_bytes: bytes,
    json_bytes_signed: bytes,
):
    """Test get_config() pass-through for signed and unsigned configurations."""
    assert image_config.get_config() == json_bytes
    assert image_config_signed.get_config() == json_bytes_signed


def test_get_config_canonical(
    image_config: ImageConfig,
    image_config_signed: ImageConfig,
    json_bytes_canonical: bytes,
    json_bytes_signed_canonical: bytes,
):
    """Test the canonical form for signed and unsigned configurations."""
    assert image_config.get_config_canonical() == json_bytes_canonical
    assert image_config_signed.get_config_canonical() == json_bytes_signed_canonical


def test_get_config_digest(
    image_config: ImageConfig,
    image_config_signed: ImageConfig,
    config_digest: str,
    config_digest_signed: str,
):
    """Test digest calculation for signed and unsigned configurations."""
    assert image_config.get_config_digest() == config_digest
    assert image_config_signed.get_config_digest() == config_digest_signed


def test_get_config_digest_canonical(
    image_config: ImageConfig,
    image_config_signed: ImageConfig,
    config_digest_canonical: str,
    config_digest_signed_canonical: str,
):
    """Test canonical digest calculation for signed and unsigned configurations."""
    assert image_config.get_config_digest_canonical() == config_digest_canonical
    assert (
        image_config_signed.get_config_digest_canonical()
        == config_digest_signed_canonical
    )


def test_get_image_layers(
    image_config: ImageConfig, image_config_signed: ImageConfig, image_layers: List
):
    """Test image layer preservation for signed and unsigned configurations."""
    assert image_config.get_image_layers() == image_layers
    assert image_config_signed.get_image_layers() == image_layers


def test_get_signature_list(
    image_config: ImageConfig,
    image_config_signed: ImageConfig,
    config_digest_canonical: str,
    signature: str,
):
    """Test signature data parsing for signed and unsigned configurations."""
    signatures_signed = image_config_signed.get_signature_list()
    assert len(signatures_signed) == 1
    assert signatures_signed[0]["digest"] == config_digest_canonical
    assert signatures_signed[0]["signature"] == signature

    signatures_unsigned = image_config.get_signature_list()
    assert not signatures_unsigned


def test_sign(
    image_config: ImageConfig,
    image_config_signed: ImageConfig,
    config_digest_canonical: str,
    signature: str,
):
    """Test configuration signing for signed and unsigned configurations."""

    signer = FakeSigner()
    assert image_config.sign(signer) == signer.signature_value
    assert image_config_signed.sign(signer) == signer.signature_value

    # Previously unsigned configurations should now contain the new signature.
    assert b"BEGIN FAKE SIGNATURE" in image_config.get_config()
    signatures = image_config.get_signature_list()
    assert len(signatures) == 1
    assert signatures[0]["digest"] == config_digest_canonical
    assert signatures[0]["signature"] == signer.signature_value

    # Previously signed configurations should now contain the original signature(s) and the new signature.
    assert b"BEGIN FAKE SIGNATURE" in image_config_signed.get_config()
    assert b"BEGIN PGP SIGNATURE" in image_config_signed.get_config()
    signatures_signed = image_config_signed.get_signature_list()
    assert len(signatures_signed) == 2
    assert signatures_signed[0]["digest"] == config_digest_canonical
    assert signatures_signed[0]["signature"] == signature
    assert signatures_signed[1]["digest"] == config_digest_canonical
    assert signatures_signed[1]["signature"] == signer.signature_value


def test_sign_endorse(
    image_config: ImageConfig,
    image_config_signed: ImageConfig,
    config_digest_canonical: str,
    config_digest_signed_canonical: str,
    signature: str,
):
    """Test configuration endorsement for signed and unsigned configurations."""

    signer = FakeSigner()
    assert image_config.sign(signer, True) == signer.signature_value
    assert image_config_signed.sign(signer, True) == signer.signature_value

    # Previously unsigned configurations should now contain the new signature.
    assert b"BEGIN FAKE SIGNATURE" in image_config.get_config()
    signatures = image_config.get_signature_list()
    assert len(signatures) == 1
    assert signatures[0]["digest"] == config_digest_canonical
    assert signatures[0]["signature"] == signer.signature_value

    # Previously signed configurations should now contain the original signature(s) and the new signature.
    assert b"BEGIN FAKE SIGNATURE" in image_config_signed.get_config()
    assert b"BEGIN PGP SIGNATURE" in image_config_signed.get_config()
    signatures_signed = image_config_signed.get_signature_list()
    assert len(signatures_signed) == 2
    assert signatures_signed[0]["digest"] == config_digest_canonical
    assert signatures_signed[0]["signature"] == signature
    assert signatures_signed[1]["digest"] == config_digest_signed_canonical
    assert signatures_signed[1]["signature"] == signer.signature_value


def test_sign_endorse_recursive(image_config: ImageConfig):
    """Test interlaced signatures and endorsements."""

    # Stack representation of a binary tree
    stack = [copy.deepcopy(image_config)]
    iterations = 6
    # Breadth first traversal ...
    for i in range(iterations):
        for _ in range(len(stack)):
            # Validate the signature / endorsement permutations of the first entry on the stack ...
            signatures = stack[0].get_signature_list()
            assert len(signatures) == i
            for sig, signature in enumerate(signatures):
                if "Signing" in signature["signature"] or sig == 0:
                    # Signature digests should be independent of the number of signatures
                    assert (
                        signature["digest"]
                        == image_config.get_config_digest_canonical()
                    )
                else:
                    # Endorsement digests should include all entities of a lower order
                    temp = copy.deepcopy(stack[0])
                    temp.set_signature_list(temp.get_signature_list()[:sig])
                    assert signature["digest"] == temp.get_config_digest_canonical()

            def append_new_image_config(
                config: ImageConfig, endorse: bool = False, iteration=i
            ):
                signer = FakeSigner(
                    "<<< {0} {1}: {2} >>>".format(
                        iteration,
                        "Endorsing" if endorse else "Signing",
                        config.get_config_digest_canonical(),
                    )
                )
                config.sign(signer, endorse)
                stack.append(config)

            # TODO: Add optimization to stop appending to the stack if they will never be validated

            # Push two more image configurations on to the stack: one signed, one endorsed ...
            append_new_image_config(copy.deepcopy(stack[0]))
            append_new_image_config(stack.pop(0), True)


def test_verify_signatures(image_config: ImageConfig):
    """Test signature verification for signed and unsigned configurations."""

    # Unsigned configurations should explicitly raise an exception.
    with pytest.raises(Exception) as exception:
        image_config.verify_signatures()
    assert str(exception.value) == "Image does not contain any signatures!"

    # Sign a previously unsigned configuration, so that only the new signature type is present.
    # Note: It is not trivial to embed "known" GPG / PKI signature types, as assumptions about the
    #       test environment are difficult to make.
    image_config.sign(FakeSigner())

    # An exception should be raised if the provider for a signature type is not known
    with pytest.raises(Exception) as exception:
        image_config.verify_signatures()
    assert str(exception.value) == "Unsupported signature type!"

    # Replace the class method for resolving signature providers ...
    original_method = Signer.for_signature
    Signer.for_signature = _signer_for_signature

    # The Signer's verify() method should be invoked.
    assert image_config.verify_signatures()["results"] == [
        {"type": "fake", "valid": True}
    ]

    # Restore the original class method
    Signer.for_signature = original_method


def test_verify_signatures_manipulated_signatures(image_config: ImageConfig):
    """Test that signature verification detects manipulated signatures."""

    # Add a single signature ...
    signer = FakeSigner()
    assert image_config.sign(signer) == signer.signature_value

    # Replace the class method for resolving signature providers ...
    original_method = Signer.for_signature
    Signer.for_signature = _signer_for_signature

    # Sanity check
    assert image_config.verify_signatures()["results"][0]["valid"] is True

    # Modify the digest value of the (first) signature ...
    signatures = image_config.get_signature_list()
    temp = copy.deepcopy(signatures)
    temp[0]["digest"] = "tampertampertamper"
    image_config.set_signature_list(temp)

    # An exception should be raised if digest value from the signature does not match the canonical digest of the image
    # configuration (without any signatures).
    with pytest.raises(Exception) as exception:
        image_config.verify_signatures()
    assert str(exception.value).startswith("Image config canonical digest mismatch:")

    # Restore the unmodified signature and endorse ...
    image_config.set_signature_list(signatures)
    assert image_config.sign(signer, True) == signer.signature_value

    # Sanity check
    assert image_config.verify_signatures()["results"][0]["valid"] is True

    # Modify the digest value of the second signature ...
    signatures = image_config.get_signature_list()
    temp = copy.deepcopy(signatures)
    temp[1]["digest"] = "tampertampertamper"
    image_config.set_signature_list(temp)

    # An exception should be raised if digest value from the signature does not match the canonical digest of the image
    # configuration (including the first signature).
    with pytest.raises(Exception) as exception:
        image_config.verify_signatures()
    assert str(exception.value).startswith("Image config canonical digest mismatch:")

    # Restore the original class method
    Signer.for_signature = original_method


def test_minimal():
    """Test minimal image configuration (for non-conformant labels)k."""

    # Note: At a minimum, [Cc]onfig key must exist with non-null value
    image_config = ImageConfig(b'{"Config":{}}')
    config_digest_canonical = image_config.get_config_digest_canonical()
    signer = FakeSigner()
    assert image_config.sign(signer) == signer.signature_value

    # A signature should always be able to be added ...
    assert b"BEGIN FAKE SIGNATURE" in image_config.get_config()
    signatures = image_config.get_signature_list()
    assert len(signatures) == 1
    assert signatures[0]["digest"] == config_digest_canonical
    assert signatures[0]["signature"] == signer.signature_value
